#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Description:
Test mpwt functions which don't need a Pathway-Tools environment.
"""

import mpwt
import os


def test_create_dats_and_lisp():
    input_path = os.path.join(*['test', 'fatty_acid_beta_oxydation_I'])
    mpwt.pathologic_input.create_flats_and_lisp(input_path, False)

    genetic_pathname = os.path.join(*['test', 'fatty_acid_beta_oxydation_I', 'genetic-elements.dat'])
    organism_pathname = os.path.join(*['test', 'fatty_acid_beta_oxydation_I', 'organism-params.dat'])
    lisp_pathname = os.path.join(*['test', 'fatty_acid_beta_oxydation_I' , 'flat_files_creation.lisp'])

    genetic_string_expected = 'NAME\t\nANNOT-FILE\tfatty_acid_beta_oxydation_I.gbk\n//\n'
    organism_string_expected = 'ID\tfatty_acid_beta_oxydation_I\nSTORAGE\tFILE\nNCBI-TAXON-ID\t511145\nNAME\tEscherichia coli str. K-12 substr. MG1655\n'
    lisp_string_expected = ''';; Most files should begin by specifying that they will be interpreted\n;; within the Lisp package called ecocyc.\n(in-package :ecocyc)\n;; Select the organism PGDB as the current PGDB\n(select-organism :org-id 'fatty_acid_beta_oxydation_I)\n;; Create attribute-values files without the progression pop-up\n(let ((*progress-noter-enabled?* NIL))\n        (create-flat-files-for-current-kb))'''

    with open(genetic_pathname, 'r') as genetic_file:
        genetic_string_found = genetic_file.read()
        assert genetic_string_found == genetic_string_expected

    with open(organism_pathname, 'r') as organism_file:
        organism_string_found = organism_file.read()
        assert organism_string_found == organism_string_expected

    with open(lisp_pathname, 'r') as lisp_file:
        lisp_string_found = lisp_file.read()
        assert lisp_string_found == lisp_string_expected

    mpwt.cleaning_input('test')


def test_check_input_and_existing_pgdb():
    mpwt.remove_pgdbs('fatty_acid_beta_oxydation_icyc,fatty_acid_beta_oxydation_i_gffcyc,fatty_acid_beta_oxydation_i_pfcyc')

    run_patho_ids, run_flat_ids = mpwt.pathologic_input.check_input_and_existing_pgdb(['fatty_acid_beta_oxydation_I', 'fatty_acid_beta_oxydation_I_pf', 'fatty_acid_beta_oxydation_I_gff'], 'test', None, 1)
    assert sorted(run_patho_ids) == sorted(['fatty_acid_beta_oxydation_I', 'fatty_acid_beta_oxydation_I_pf', 'fatty_acid_beta_oxydation_I_gff'])
    assert not run_flat_ids

    run_patho_ids, run_flat_ids = mpwt.pathologic_input.check_input_and_existing_pgdb(['wrong.gbk.id.gbk'], 'test_wrong_id', None, 1)
    assert not run_patho_ids
    assert not run_flat_ids



"""
Check input folders and input files.

Create PathoLogic input files:
-organism-params.dat
-genetic-elements.dats
-flat_files_creation.lisp
"""

import csv
import logging
import os
import sys

from Bio import SeqIO
from gffutils.iterators import DataIterator
from mpwt import utils

logger = logging.getLogger(__name__)


def compare_input_ids_to_ptools_ids(compare_ids, ptools_run_ids, set_operation):
    """ Compare species IDs in input folder with IDs present in PGDB folder.

    Args:
        compare_ids (list): species IDs (folder and GBK/GFF file name)
        ptools_run_ids (list): PGDB IDs inside ptools-local
        set_operation (str): difference or intersection between the two inputs list
    Returns:
        list: ID intersecting or different between the two lists
    """
    # To compare them, lower case the species IDs, then run a difference or an intersection between set.
    lower_case_compare_ids = list(map(lambda x:x.lower(), compare_ids))

    lower_compare_ids = dict(zip(lower_case_compare_ids, compare_ids))

    # Difference to obtain all IDs which are not in PGDB folder (which need to be run on PathoLogic).
    if set_operation == 'difference':
        compare_ids_ptools = set(lower_case_compare_ids) - set(ptools_run_ids)

    # Intersection to obtain all IDs which are already in PGDB folder.
    # Which need only to create BioPAX/flat files and move them in output folder.
    elif set_operation == 'intersection':
        compare_ids_ptools = set(ptools_run_ids).intersection(set(lower_case_compare_ids))

    new_compare_ids = [lower_compare_ids[compare_id] for compare_id in compare_ids_ptools]

    return new_compare_ids


def check_input_and_existing_pgdb(run_ids, input_folder, output_folder, number_cpu_to_use):
    """ Check input structure and data in output folder and ptools-local.

    Args:
        run_ids (list): species IDs (folder and GBK/GFF file name)
        input_folder (str): pathname to the input folder
        output_folder (str): pathname to the output folder
        number_cpu_to_use (int): number of CPU to use for multiprocessing
    Returns:
        list: input IDs for PathoLogic and BioPAX/flat files creation
        list: input IDs for BioPAX/flat files creation
    """
    # Check if there are files/folders inside the input folder.
    # And remove hidden folder/file (beginning with '.').
    species_folders = [species_folder for species_folder in os.listdir(input_folder) if not species_folder.startswith('.')]
    if len(species_folders) == 0:
        logger.critical("No folder containing genbank/gff file. In {0} you must have sub-folders containing Genbank/GFF file.".format(input_folder))
        return None, None

    # Remove Pathologic taxon ID file.
    if 'taxon_id.tsv' in species_folders:
        species_folders.remove('taxon_id.tsv')

    # Check if there is a Genbank, a GFF or a PathoLogic file inside each subfolder.
    check_species_folders = []
    for species_folder in species_folders:
        species_input_files = []
        species_folder_path = os.path.join(input_folder, species_folder)
        for species_file in os.listdir(species_folder_path):
            species_filename, species_file_extension = os.path.splitext(species_file)
            if species_file_extension in ['.gbk', '.gbff', '.gff']:
                if species_filename == species_folder:
                    check_species_folders.append(species_folder)
                    species_input_files.append(species_file_extension)
            if any(input_extension in species_file for input_extension in ['.pf']):
                check_species_folders.append(species_folder)
                species_input_files.append(species_file_extension)
        species_input_files = list(set(species_input_files))
        if len(species_input_files) > 1:
            logger.critical('Multiple input files for {0}, there must be only one type of files among: GenBank, GFF or multiple PF files'.format(species_folder))
            return None, None
        elif len(species_input_files) == 0:
            logger.critical('Missing input file for {0}. A GenBank file, GFF file or multiple PF files are required.'.format(species_folder))
            return None, None

    check_species_folders = list(set(check_species_folders))

    missing_input_files = list(set(run_ids) - set(check_species_folders))
    if len(check_species_folders) == 0:
        logger.critical('Missing Genbank/GFF/PF file for: {0} \nCheck for input files (.gbk/.gbff/.gff/.pf)'.format(','.join(missing_input_files)))
        return None, None

    # Check the structure of the input folder.
    invalid_characters = ['.', '/']
    for species_folder in check_species_folders:
        species_folder_path = os.path.join(input_folder, species_folder)
        if os.path.isfile(species_folder_path):
            logger.critical('Error: file inside the input_folder ({0}) instead of a subfolder. Check that you have a structure file of input_folder/species_1/species1.gbk and not input_folder/species_1.gbk.'.format(species_folder_path))
            return None, None
        elif os.path.isdir(species_folder_path):
            if any(char in invalid_characters for char in species_folder):
                logger.critical('Error: . or / in genbank/gff name {0} \nGenbank name is used as an ID in Pathway Tools and Pathway Tools does not create PGDB with . in ID.'.format(species_folder))
                return None, None

    # Take run_ids and remove folder with error (with the intersection with check_species_folders) and if there is already present output.
    clean_run_ids = set(run_ids).intersection(set(check_species_folders))
    if output_folder:
        if os.path.exists(output_folder):
            if os.path.isdir(output_folder):
                already_present_outputs = [output_pgdb for output_pgdb in os.listdir(output_folder)]
                new_run_ids = clean_run_ids - set(already_present_outputs)
                new_run_ids = list(new_run_ids)
                for pgdb in already_present_outputs:
                    if pgdb in clean_run_ids:
                        logger.warning("! PGDB {0} already in output folder {1}, no inference will be launched on this species.".format(pgdb, output_folder))

                if len(new_run_ids) == 0:
                    logger.info("All PGDBs are already present in the output folder. Remove them if you want a new inference.")
                    return None, None
            else:
                logger.info(output_folder + " is not a valid output folder.")
                return None, None
        else:
            new_run_ids = list(clean_run_ids)

    else:
        new_run_ids = list(clean_run_ids)

    # Check for PGDB in ptools-local to see if PGDB are already present but they haven't been exported.
    already_present_pgdbs = [pgdb_species_folder[:-3] for pgdb_species_folder in utils.list_pgdb()]

    # Check the already finished PGDBs.
    if already_present_pgdbs:
        pathologic_builds = compare_input_ids_to_ptools_ids(new_run_ids, already_present_pgdbs, 'intersection')

        # Check for unfinished build of PGDB using their pathologic.log file.
        logger.info("Check and delete unfinished builds of Pathway Tools.")
        unfinished_builds = []
        finished_builds = []
        for pathologic_build in pathologic_builds:
            pathologic_build_lower = pathologic_build.lower()
            pathologic_file = os.path.join(*[input_folder, pathologic_build, 'pathologic.log'])
            if os.path.exists(pathologic_file):
                with open(pathologic_file, 'r') as pathologic_log:
                    pathologic_string = pathologic_log.read()
                    if 'Done' in pathologic_string:
                        finished_builds.append(pathologic_build_lower)
                    else:
                        unfinished_builds.append(pathologic_build_lower)

        # Delete the unfinished PGDBs.
        if unfinished_builds:
            utils.remove_pgdbs([unfinished_build + 'cyc' for unfinished_build in unfinished_builds], number_cpu_to_use)

        already_present_pgdbs = list(set(already_present_pgdbs) - set(unfinished_builds))

        run_patho_flat_ids = compare_input_ids_to_ptools_ids(new_run_ids, already_present_pgdbs, 'difference')
        run_flat_ids = compare_input_ids_to_ptools_ids(new_run_ids, already_present_pgdbs, 'intersection')

        for run_flat_id in run_flat_ids:
            logger.info("! PGDB {0} already in ptools-local, no PathoLogic inference will be launched on this species.".format(run_flat_id))
        return run_patho_flat_ids, run_flat_ids

    return new_run_ids, None


def create_flat_creation_script(pgdb_id, lisp_pathname):
    """ Create a lisp script allowing flat files creation.

    Args:
        pgdb_id (str): ID of a PGDB
        lisp_pathname (str): pathname to the output list script
    Returns:
        bool: True if lisp_pathname has been created
    """
    with open(lisp_pathname, 'w', encoding='utf-8') as lisp_file:
        lisp_file.write(';; Most files should begin by specifying that they will be interpreted')
        lisp_file.write('\n')
        lisp_file.write(';; within the Lisp package called ecocyc.')
        lisp_file.write('\n')
        lisp_file.write('(in-package :ecocyc)')
        lisp_file.write('\n')
        lisp_file.write(';; Select the organism PGDB as the current PGDB')
        lisp_file.write('\n')
        lisp_file.write("(select-organism :org-id '" + pgdb_id + ')')
        lisp_file.write('\n')
        lisp_file.write(';; Create attribute-values files without the progression pop-up')
        lisp_file.write('\n')
        lisp_file.write('(let ((*progress-noter-enabled?* NIL))')
        lisp_file.write('\n')
        lisp_file.write('        (create-flat-files-for-current-kb))')

    return os.path.isfile(lisp_pathname)


def extract_taxon_id(run_folder, pgdb_id, taxon_id, taxon_file):
    """ Extract taxon ID from taxon_id.tsv file.

    Args:
        run_folder (str): ID of a species of the input folder
        pgdb_id (str): ID of a PGDB
        taxon_id (str): Taxon ID for the corresponding species
        taxon_file (bool): Boolean indicating if a taxon_file must be used
    Returns:
        taxon_error (bool): Error status (True: error, False: no error).
        taxon_id (str): Taxon ID for the corresponding species
        taxon_datas (dict): Name of element file (or 'one_input' if only one file)
    """
    input_folder =  os.path.abspath(os.path.join(run_folder ,os.pardir))
    taxon_datas = {}
    taxon_id_found = None

    known_element_types = [':CHRSM', ':PLASMID', ':MT', ':PT', ':CONTIG']
    known_codon_table = ['0', '1', '2', '3', '4', '5', '6', '9', '10', '11', '12', '13', '14', '15', '16', '21', '22', '23']
    known_species = []

    taxon_id_path = os.path.join(input_folder, 'taxon_id.tsv')
    if not os.path.exists(taxon_id_path):
        logger.critical('Missing taxon_id.tsv file in {0}.'.format(input_folder))
        return True, None, None

    with open(taxon_id_path) as taxon_id_file:
        taxon_id_reader = csv.DictReader(taxon_id_file, delimiter='\t')
        for data in taxon_id_reader:
            if 'species' not in data:
                logger.critical('Missing "species" header in taxon_id.tsv file {0}.'.format(taxon_id_path))
                return True, None, None
            species = data['species']
            known_species.append(species)
            if pgdb_id == species:
                if 'taxon_id' in data:
                    if data['taxon_id'] != '':
                        if not taxon_id_found:
                            taxon_id = data['taxon_id']
                            taxon_id_found = True
                            logger.info('taxon_id.tsv: find taxon ID {0} for {1}'.format(taxon_id, pgdb_id))
                    else:
                        logger.critical('Missing taxon ID for {0} in {1}.'.format(pgdb_id, taxon_id_path))
                        return True, None, None
                else:
                    logger.critical('Missing taxon ID for {0} in {1}.'.format(pgdb_id, taxon_id_path))
                    return True, None, None

                if 'circular' in data:
                    if data['circular'] != '':
                        if data['circular'] == 'Y' or data['circular'] == 'N':
                            circular = data['circular']
                        else:
                            logger.critical('taxon_id.tsv: wrong circular for {0}, {1} instead of Y or N'.format(pgdb_id, data['circular']))
                            return True, None, None
                    else:
                        circular = None
                else:
                    circular = None

                if 'element_type' in data:
                    if data['element_type'] != '':
                        if data['element_type'] in known_element_types:
                            element_type = data['element_type']
                        else:
                            logger.critical('taxon_id.tsv: wrong element_type for {0}, {1} instead of {2}'.format(pgdb_id, data['element_type'], ', '.join(known_element_types)))
                            return True, None, None
                    else:
                        element_type = None
                else:
                    element_type = None

                if 'codon_table' in data:
                    if data['codon_table'] != '':
                        if data['codon_table'] in known_codon_table:
                            codon_table = data['codon_table']
                        else:
                            logger.critical('taxon_id.tsv: wrong codon_table for {0}, {1} instead of {2}'.format(pgdb_id, data['codon_table'], ', '.join(known_codon_table)))
                            return True, None, None
                    else:
                        codon_table = None
                else:
                    codon_table = None

                if 'corresponding_file' in data:
                    if data['corresponding_file'] != '':
                        corresponding_file = data['corresponding_file']
                        taxon_datas[corresponding_file] = {}
                        if circular is not None:
                            taxon_datas[corresponding_file]['circular'] = circular
                        if element_type is not None:
                            taxon_datas[corresponding_file]['element_type'] = element_type
                        if codon_table is not None:
                            taxon_datas[corresponding_file]['codon_table'] = codon_table
                    else:
                        if circular is not None:
                            taxon_datas['circular'] = circular
                        if element_type is not None:
                            taxon_datas['element_type'] = element_type
                        if codon_table is not None:
                            taxon_datas['codon_table'] = codon_table
                else:
                    if circular is not None:
                        taxon_datas['circular'] = circular
                    if element_type is not None:
                        taxon_datas['element_type'] = element_type
                    if codon_table is not None:
                        taxon_datas['codon_table'] = codon_table

    if pgdb_id not in known_species and taxon_id == '':
        logger.critical('Missing pgdb ID for {0} in {1}.'.format(pgdb_id, taxon_id_path))
        return True, None, None

    if taxon_file and pgdb_id not in known_species:
        logger.critical('Missing pgdb ID for {0} in {1}.'.format(pgdb_id, taxon_id_path))
        return True, None, None

    return False, taxon_id, taxon_datas


def create_flats_and_lisp(run_folder, taxon_file):
    """
    Read Genbank/GFF/PF files and create Pathway Tools needed file.
    Create also a lisp file to create flat files from Pathway tools results.
    The name of the PGDB created by Pathway Tools will be the name of the species with '_' instead of space.

    Create organism-params.dat:
    ID  pgdb_id
    STORAGE FILE
    NCBI-TAXON-ID   taxon_id
    NAME    species_name

    Create genetic-elements.dats:
    NAME    
    ANNOT-FILE  gbk_name
    //

    Create flat_files_creation.lisp:
    (in-package :ecocyc)
    (select-organism :org-id 'pgdb_id)
    (create-flat-files-for-current-kb)

    Args:
        run_folder (str): ID of a species of the input folder
    Returns:
        list: boolean list, True if all files have been created
   """
    # Look for a Genbank/GFF files in the run folder.
    # PGDB ID corresponds to the name of the species folder.
    pgdb_id = os.path.basename(run_folder)
    gbk_name = pgdb_id + ".gbk"
    gbk_pathname = os.path.join(run_folder, gbk_name)
    gbff_name = pgdb_id + ".gbff"
    gbff_pathname = os.path.join(run_folder, gbff_name)
    gff_name = pgdb_id + ".gff"
    gff_pathname = os.path.join(run_folder, gff_name)

    organism_dat = os.path.join(run_folder, 'organism-params.dat')
    genetic_dat = os.path.join(run_folder, 'genetic-elements.dat')
    lisp_pathname = os.path.join(run_folder, 'flat_files_creation.lisp')

    fasta_extensions = ['.fasta', '.fsa']

    taxon_id = ""
    taxon_error = False
    species_name = ""
    taxon_datas = {}

    if os.path.isfile(gbk_pathname) or os.path.isfile(gbff_pathname):
        if os.path.isfile(gbk_pathname):
            input_name = gbk_name
            input_path = gbk_pathname
        else:
            input_name = gbff_name
            input_path = gbff_pathname
        # Take the species name and the taxon id from the genbank file.
        with open(input_path, "r") as gbk:
            # Take the first record of the genbank (first contig/chromosome) to retrieve the species name.
            try:
                first_seq_record = next(SeqIO.parse(gbk, "genbank"))
            except StopIteration:
                logger.critical('Issue with the genbank {0}, it can be empty or malformatted.'.format(input_path))
                return None

            try:
                species_name = first_seq_record.annotations['organism']
            except KeyError:
                logger.critical('No organism in the Genbank {0} In the SOURCE you must have: ORGANISM  Species name'.format(pgdb_id))
                return None

            # Take the source feature of the first record.
            # This feature contains the taxon ID in the db_xref qualifier.
            src_features = [feature for feature in first_seq_record.features if feature.type == "source"]
            for src_feature in src_features:
                if 'db_xref' in src_feature.qualifiers:
                    src_dbxref_qualifiers = src_feature.qualifiers['db_xref']
                    for src_dbxref_qualifier in src_dbxref_qualifiers:
                        if 'taxon:' in src_dbxref_qualifier:
                            taxon_id = src_dbxref_qualifier.replace('taxon:', '')
                if not taxon_id:
                    logger.info('No taxon ID in the Genbank {0} In the FEATURES source you must have: /db_xref="taxon:taxonid" Where taxonid is the Id of your organism. You can find it on the NCBI.'.format(gbk_pathname))
                    logger.info('Try to look in the taxon_id.tsv file')
                    taxon_error, taxon_id, taxon_datas = extract_taxon_id(run_folder, pgdb_id, taxon_id, taxon_file)
            if taxon_file:
                taxon_error, taxon_id, taxon_datas = extract_taxon_id(run_folder, pgdb_id, taxon_id, taxon_file)

    elif os.path.isfile(gff_pathname):
        input_name = gff_name
        # Check if there is a fasta file.
        gff_fasta = None
        for fasta_extension in fasta_extensions:
            fasta_input_name = input_name.replace('.gff', fasta_extension)
            fasta_path = os.path.join(run_folder, fasta_input_name)
            if os.path.exists(fasta_path):
                gff_fasta = fasta_input_name
        if not gff_fasta:
            logger.critical('No fasta file (.fasta or .fsa) with the GFF of {0}'.format(pgdb_id))
            return None

        # Instead of parsing and creating a database from the GFF, parse the file and extract the first region feature.
        try:
            region_feature = [feature for feature in DataIterator(gff_pathname) if feature.featuretype == 'region'][0]
        except IndexError:
            logger.critical('No region feature in the GFF file of {0}, GFF file must have region features.'.format(pgdb_id))
            return None

        try:
            region_feature.attributes['Dbxref']
        except KeyError:
            logger.critical('No Dbxref in GFF file of {0} GFF file must have a ;Dbxref=taxon:taxonid; in the region feature.'.format(pgdb_id))

        for dbxref in region_feature.attributes['Dbxref']:
            if 'taxon' in dbxref:
                taxon_id = dbxref.split('taxon:')[1]
        if not taxon_id or taxon_file:
            if not taxon_id:
                logger.info('Missing "taxon:" in GFF file of {0} GFF file must have a ;Dbxref=taxon:taxonid; in the region feature.'.format(pgdb_id))
                logger.info('Try to look in the taxon_id.tsv file')
            taxon_error, taxon_id, taxon_datas = extract_taxon_id(run_folder, pgdb_id, taxon_id, taxon_file)

    # Look for PF files.
    elif all([True for species_file in os.listdir(run_folder) if '.pf' in species_file or '.fasta' in species_file or '.fsa' in species_file]):
        for species_file in os.listdir(run_folder):
            if '.pf' in species_file:
                # Check if there is a fasta file.
                pf_fasta = None
                for fasta_extension in fasta_extensions:
                    fasta_species_name = species_file.replace('.pf', fasta_extension)
                    fasta_path = os.path.join(run_folder, fasta_species_name)
                    if os.path.exists(fasta_path):
                        pf_fasta = fasta_species_name
                if not pf_fasta:
                    logger.critical('No fasta file (.fasta or .fsa) with the Pathologic file of {0}'.format(pgdb_id))
                    return None

        taxon_error, taxon_id, taxon_datas = extract_taxon_id(run_folder, pgdb_id, taxon_id, taxon_file)

    if taxon_error == True:
        logger.critical('Issue with taxon ID of {0}.'.format(run_folder))
        return None

    # Create the organism-params dat file.
    with open(organism_dat, 'w', encoding='utf-8') as organism_file:
        organism_writer = csv.writer(organism_file, delimiter='\t', lineterminator='\n')
        organism_writer.writerow(['ID', pgdb_id])
        organism_writer.writerow(['STORAGE', "FILE"])
        organism_writer.writerow(['NCBI-TAXON-ID', taxon_id])
        organism_writer.writerow(['NAME', species_name])

    # Create the genetic-elements dat file.
    with open(genetic_dat, 'w', encoding='utf-8') as genetic_file:
        if os.path.isfile(gff_pathname) or os.path.isfile(gbk_pathname) or os.path.isfile(gbff_pathname):
            genetic_writer = csv.writer(genetic_file, delimiter='\t', lineterminator='\n')
            genetic_writer.writerow(['NAME', ''])
            genetic_writer.writerow(['ANNOT-FILE', input_name])
            if os.path.isfile(gff_pathname):
                genetic_writer.writerow(['SEQ-FILE', gff_fasta])
            if 'circular' in taxon_datas:
                circular = taxon_datas['circular']
                genetic_writer.writerow(['CIRCULAR?', circular])
            if 'element_type' in taxon_datas:
                element_type = taxon_datas['element_type']
                genetic_writer.writerow(['TYPE', element_type])
            if 'codon_table' in taxon_datas:
                codon_table = taxon_datas['codon_table']
                genetic_writer.writerow(['CODON-TABLE', codon_table])
            genetic_writer.writerow(['//'])
        elif all([True for species_file in os.listdir(run_folder) if '.pf' in species_file or '.fasta' in species_file or '.fsa' in species_file]):
            genetic_writer = csv.writer(genetic_file, delimiter='\t', lineterminator='\n')
            for species_file in os.listdir(run_folder):
                    if '.pf' in species_file:
                        species_file_name = os.path.splitext(species_file)[0]
                        genetic_writer.writerow(['NAME', species_file.replace('.pf', '')])
                        genetic_writer.writerow(['ID', species_file.replace('.pf', '')])
                        genetic_writer.writerow(['ANNOT-FILE', species_file])
                        fasta_path = os.path.join(run_folder, species_file.replace('.pf', '.fasta'))
                        fsa_path = os.path.join(run_folder, species_file.replace('.pf', '.fsa'))
                        if os.path.exists(fasta_path):
                            genetic_writer.writerow(['SEQ-FILE', species_file.replace('.pf', '.fasta')])
                        elif os.path.exists(fsa_path):
                            genetic_writer.writerow(['SEQ-FILE', species_file.replace('.pf', '.fsa')])

                        if species_file_name in taxon_datas:
                            if 'circular' in taxon_datas[species_file_name]:
                                circular = taxon_datas[species_file_name]['circular']
                                genetic_writer.writerow(['CIRCULAR?', circular])
                            if 'element_type' in taxon_datas[species_file_name]:
                                element_type = taxon_datas[species_file_name]['element_type']
                                genetic_writer.writerow(['TYPE', element_type])
                            if 'codon_table' in taxon_datas[species_file_name]:
                                codon_table = taxon_datas[species_file_name]['codon_table']
                                genetic_writer.writerow(['CODON-TABLE', codon_table])
                        else:
                            if 'circular' in taxon_datas:
                                circular = taxon_datas['circular']
                                genetic_writer.writerow(['CIRCULAR?', circular])
                            if 'element_type' in taxon_datas:
                                element_type = taxon_datas['element_type']
                                genetic_writer.writerow(['TYPE', element_type])
                            if 'codon_table' in taxon_datas:
                                codon_table = taxon_datas['codon_table']
                                genetic_writer.writerow(['CODON-TABLE', codon_table])
                        genetic_writer.writerow(['//'])
    # Create the lisp script.
    check_lisp_file = create_flat_creation_script(pgdb_id, lisp_pathname)

    return all([os.path.isfile(organism_dat), os.path.isfile(genetic_dat), check_lisp_file])


def read_taxon_id(run_folder):
    taxon_ids = {}

    for input_folder in os.listdir(run_folder):
        input_folder_path = os.path.join(run_folder, input_folder)
        for input_file in os.listdir(input_folder_path):
            if '.gbk' in input_file:
                gbk_pathname = os.path.join(input_folder_path, input_file)
                # Take the species name and the taxon id from the genbank file.
                with open(gbk_pathname, "r") as gbk:
                    # Take the first record of the genbank (first contig/chromosome) to retrieve the species name.
                    first_seq_record = next(SeqIO.parse(gbk, "genbank"))
                    # Take the source feature of the first record.
                    # This feature contains the taxon ID in the db_xref qualifier.
                    src_features = [feature for feature in first_seq_record.features if feature.type == "source"]
                    for src_feature in src_features:
                        try:
                            src_dbxref_qualifiers = src_feature.qualifiers['db_xref']
                            for src_dbxref_qualifier in src_dbxref_qualifiers:
                                if 'taxon:' in src_dbxref_qualifier:
                                    taxon_id = src_dbxref_qualifier.replace('taxon:', '')
                        except KeyError:
                            logger.info('No taxon ID in the Genbank {0} In the FEATURES source you must have: /db_xref="taxon:taxonid" Where taxonid is the Id of your organism. You can find it on the NCBI.'.format(gbk_pathname))

            elif '.gff' in input_file:
                gff_pathname = os.path.join(input_folder_path, input_file)

                # Instead of parsing and creating a database from the GFF, parse the file and extract the first region feature.
                try:
                    region_feature = [feature for feature in DataIterator(gff_pathname) if feature.featuretype == 'region'][0]
                except IndexError:
                    raise IndexError('No region feature in the GFF file of {0}, GFF file must have region features.'.format(input_folder))

                try:
                    region_feature.attributes['Dbxref']
                except KeyError:
                    raise KeyError('No Dbxref in GFF file of {0} GFF file must have a ;Dbxref=taxon:taxonid; in the region feature.'.format(input_folder))

                for dbxref in region_feature.attributes['Dbxref']:
                    if 'taxon' in dbxref:
                        taxon_id = dbxref.split('taxon:')[1]

            elif '.pf' in input_file:
                logger.info('No taxon ID associated to a PathoLogic Format. {0} will have a missing taxon_id'.format(input_folder))
                taxon_id = "missing"
        taxon_ids[input_folder] = taxon_id

    return taxon_ids


def pwt_input_files(run_folder, taxon_file):
    """
    Check if files needed by Pathway Tools are available, if not create them.
    Check if there is a pathologic.log from a previous run. If yes, delete it.

    Args:
        run_folder (str): path to the input folder
        taxon_file (str): path to the taxon_id.tsv file
    """
    required_files = set(['organism-params.dat', 'genetic-elements.dat', 'flat_files_creation.lisp'])
    files_in = set(next(os.walk(run_folder))[2])

    species_folder = os.path.basename(run_folder)
    pathologic_log = os.path.join(run_folder, 'pathologic.log')

    if 'pathologic.log' in files_in:
        os.remove(pathologic_log)

    error_found = False
    missing_string = ''
    if required_files.issubset(files_in):
        missing_string = 'no missing files'
    else:
        missing_string = 'missing {0}'.format('; '.join(required_files.difference(files_in))) + '. Inputs file created for {0}'.format(species_folder)
        check_datas_lisp = create_flats_and_lisp(run_folder, taxon_file)
        if check_datas_lisp is None:
            logger.critical('Error with the creation of input files of {0}.'.format(run_folder))
            error_found = True
            return error_found

    logger.info('Checking inputs for {0}: {1}.'.format(species_folder, missing_string))

    return error_found


def create_only_flat_lisp(pgdbs_folder_path, tmp_folder):
    """
    Create a lisp script file for each PGDB in the ptools-local folder.
    Return a generator with the PGDB IDs.

    Args:
        pgdbs_folder_path (str): pathname to species PGDB in ptools-local
        tmp_folder (str): termporary folder where lisp script will be stored
    Returns:
        generator: generator with the PGDB IDs.
    """
    for species_pgdb in os.listdir(pgdbs_folder_path):
        species_dir_path = os.path.join(pgdbs_folder_path, species_pgdb)
        if os.path.isdir(species_dir_path):
            pgdb_id = species_pgdb[:-3]
            pgdb_pathname = os.path.join(tmp_folder, pgdb_id)
            tmp_pgdb_path = os.path.join(tmp_folder, pgdb_id)
            os.mkdir(tmp_pgdb_path)
            lisp_pathname = os.path.join(pgdb_pathname, 'flat_files_creation.lisp')
            check_lisp_file = create_flat_creation_script(pgdb_id, lisp_pathname)
            if not check_lisp_file:
                raise Exception('Error with the creation of the lisp script for {0}'.format(species_pgdb))

            yield pgdb_id



"""
Workflow of mpwt:
    -create input files (pathologic_input)
    -launch Pathway Tools in multiprocess (pwt_wrapper)
    -check the results (results_check)
"""

import csv
import logging
import os
import shutil
import sys
import time

from mpwt import utils
from mpwt.pwt_wrapper import run_pwt, run_pwt_flat, run_move_pgdb
from mpwt.results_check import check_dat, check_pwt
from mpwt.pathologic_input import check_input_and_existing_pgdb, pwt_input_files, create_only_flat_lisp, create_flat_creation_script, read_taxon_id
from multiprocessing import Pool

logger = logging.getLogger(__name__)


def multiprocess_pwt(input_folder=None, output_folder=None, patho_inference=None,
                     patho_hole_filler=None, patho_operon_predictor=None,
                     patho_transporter_inference=None, no_download_articles=None,
                     flat_creation=None, dat_extraction=None, xml_extraction=None,
                     owl_extraction=None, col_extraction=None, size_reduction=None,
                     number_cpu=None, patho_log=None, ignore_error=None,
                     pathway_score=None, taxon_file=None, verbose=None):
    """
    Function managing all the workflow (from the creatin of the input files to the results).
    Use it when you import mpwt in a script.

    Args:
        input_folder (str): pathname to input folder
        output_folder (str): pathname to output folder
        patho_inference (bool): PathoLogic inference (True/False)
        patho_hole_filler (bool): PathoLogic hole filler (True/False)
        patho_operon_predictor (bool): PathoLogic operon predictor (True/False)
        patho_transporter_inference (bool): PathoLogic Transport Inference Parser (True/False)
        no_download_articles (bool): turning off loading of PubMed citations (True/False)
        flat_creation (bool): BioPAX/attributes-values flat files creation (True/False)
        dat_extraction (bool): move BioPAX/attributes-values files to output folder (True/False)
        xml_extraction (bool): move metabolic-reactions.xml to output folder (True/False)
        owl_extraction (bool): move owl files to output folder (True/False)
        col_extraction (bool): move tabular files to output folder (True/False)
        size_reduction (bool): delete ptools-local data at the end (True/False)
        number_cpu (int): number of CPU used (default=1)
        patho_log (str): pathname to mpwt log folder
        ignore_error (bool): Ignore error during PathoLogic inference (True/False)
        pathway_score (float): score between 0 and 1 to accept or reject pathway
        taxon_file (str): pathname to the mpwt taxon ID file
        verbose (bool): verbose argument
    """
    if verbose:
        logger.setLevel(logging.DEBUG)

    start_time = time.time()
    times = []
    steps = []
    times.append(start_time)
    steps.append('start')

    # Check if Pathway Tools is in the path.
    # Find PGDB folder path.
    ptools_local_path = utils.find_ptools_path()
    pgdbs_folder_path = os.path.join(*[ptools_local_path, 'pgdbs', 'user'])

    # Check if ptools-local is accessible.
    error = utils.check_ptools_local_pwt()
    if error:
        sys.exit(1)

    # Check if patho_hole_filler or patho_log are launched with patho_inference.
    if (patho_hole_filler and not patho_inference) or (patho_log and not patho_inference):
        sys.exit('To use either --hf/patho_hole_filler or --log/patho_log, you need to add the --patho/patho_inference argument.')

    # Check if size_reduction is used with output_folder.
    if size_reduction and not output_folder:
        sys.exit('To use -r/size_reduction, you need to give an output folder (-o/output_folder).')

    # Check if ignore_error is used with patho_inference.
    if ignore_error and not patho_inference:
        sys.exit('To use --ignore-error/ignore_error, you need to use the --patho/patho_inference argument.')

    # Check if taxon_file is used with patho_inference.
    if (taxon_file and not patho_inference) and (taxon_file and not input_folder):
        sys.exit('To use --taxon-file/taxon_file, you need to use the --patho/patho_inference argument. Or you can use it with the -f argument to create the taxon file from data.')

    #Check if patho_operon_predictor is used with patho_inference.
    if patho_operon_predictor and not patho_inference:
        sys.exit('To use --op/patho_operon_predictor, you need to use the --patho/patho_inference argument.')

    #Check if no_download_articles is used with patho_inference.
    if no_download_articles and not patho_inference:
        sys.exit('To use --nc/no_download_articles, you need to use the --patho/patho_inference argument.')

    #Check if patho_transporter_inference is used with patho_inference.
    if patho_transporter_inference and not patho_inference:
        sys.exit('To use --tp/patho_transporter_inference, you need to use the --patho/patho_inference argument.')

    #Check if no_download_articles is used with patho_inference.
    if pathway_score and not patho_inference:
        sys.exit('To use -p/pathway_score, you need to use the --patho/patho_inference argument.')

    #Check if no_download_articles is used with output_folder.
    if dat_extraction and not output_folder:
        sys.exit('To use --md/dat_extraction, you need to use the -o/output_folder argument.')

    #Check if xml_extraction is used with output_folder.
    if xml_extraction and not output_folder:
        sys.exit('To use --mx/xml_extraction, you need to use the -o/output_folder argument.')

    #Check if owl_extraction is used with output_folder.
    if owl_extraction and not output_folder:
        sys.exit('To use --mo/owl_extraction, you need to use the -o/output_folder argument.')

    #Check if col_extraction is used with output_folder.
    if col_extraction and not output_folder:
        sys.exit('To use --mc/col_extraction, you need to use the -o/output_folder argument.')

    # Use the number of cpu given by the user or 1 CPU.
    if number_cpu:
        try:
            number_cpu_to_use = int(number_cpu)
        except ValueError:
            raise ValueError('The number of CPU must be an integer.')
    else:
        number_cpu_to_use = 1
    mpwt_pool = Pool(processes=number_cpu_to_use)

    if input_folder:
        if not os.path.exists(input_folder):
            logger.critical('mpwt can not run: ' + input_folder + ' does not exist.')
            return
        if not os.path.isdir(input_folder):
            logger.critical('mpwt can not run: ' + input_folder + ' is not a directory.')
            return

    # Create taxon file in the input folder.
    if taxon_file and input_folder and not patho_inference:
        taxon_file_pathname = os.path.join(input_folder, 'taxon_id.tsv')
        if os.path.exists(taxon_file_pathname):
            sys.exit('taxon ID file (' + taxon_file_pathname + ') already exists.')
        else:
            taxon_ids = read_taxon_id(input_folder)
            with open(taxon_file_pathname, 'w', encoding='utf-8') as taxon_id_file:
                taxon_id_writer = csv.writer(taxon_id_file, delimiter='\t')
                taxon_id_writer.writerow(['species', 'taxon_id'])
                for species, taxon_id in taxon_ids.items():
                    taxon_id_writer.writerow([species, taxon_id])

    # Turn off loading of pubmed entries.
    if no_download_articles:
        utils.pubmed_citations(activate_citations=False)

    # Modify pathway prediction score.
    if pathway_score:
        utils.modify_pathway_score(pathway_score)

    # Check input folder and create input files for PathoLogic.
    if input_folder:
        run_ids = [folder_id for folder_id in next(os.walk(input_folder))[1]]
        run_patho_flat_ids, run_flat_ids = check_input_and_existing_pgdb(run_ids, input_folder, output_folder, number_cpu_to_use)

        # Launch PathoLogic inference on species with no PGDBs.
        if run_patho_flat_ids:
            # Create the list containing all the data used by the multiprocessing call.
            multiprocess_pwt_input_files = []
            multiprocess_run_pwts = []
            multiprocess_run_pwt_flats = []
            multiprocess_run_move_pgdbs = []
            for run_patho_flat_id in run_patho_flat_ids:
                input_folder_path = os.path.join(input_folder, run_patho_flat_id)
                species_pgdb_folder = os.path.join(pgdbs_folder_path, run_patho_flat_id.lower() + 'cyc')
                input_run_move_pgdbs = [run_patho_flat_id, species_pgdb_folder]
                input_run_move_pgdbs.extend([dat_extraction, output_folder, size_reduction, xml_extraction, owl_extraction, col_extraction])
                multiprocess_pwt_input_files.append([input_folder_path, taxon_file])
                multiprocess_run_pwts.append([input_folder_path, patho_hole_filler, patho_operon_predictor, patho_transporter_inference])
                multiprocess_run_pwt_flats.append([input_folder_path])
                multiprocess_run_move_pgdbs.append(input_run_move_pgdbs)

            logger.info('~~~~~~~~~~Creation of input data from Genbank/GFF/PF~~~~~~~~~~')
            input_error_status = mpwt_pool.starmap(pwt_input_files, multiprocess_pwt_input_files)
            if any(input_error_status):
                close_mpwt(mpwt_pool, no_download_articles, pathway_score)
                sys.exit('Error during PathoLogic input files creation.')

            input_time = time.time()
            times.append(input_time)
            steps.append('pwt input creation')
            logger.info('----------End of creation of input data from Genbank/GFF/PF: {0:.2f}s----------'.format(times[-1] - times[-2]))

            # Launch PathoLogic.
            if patho_inference:
                logger.info('~~~~~~~~~~Inference on the data~~~~~~~~~~')
                error_status = mpwt_pool.starmap(run_pwt, multiprocess_run_pwts)

                # Check PathoLogic build.
                logger.info('~~~~~~~~~~Check inference~~~~~~~~~~')
                passed_inferences = check_pwt(multiprocess_run_pwts, patho_log)
                if any(error_status):
                    if ignore_error:
                        logger.critical('Error during inference. Process stopped. Look at the command log. Also by using --log argument, you can have additional information.')
                    else:
                        close_mpwt(mpwt_pool, no_download_articles, pathway_score)
                        sys.exit('Error during inference. Process stopped. Look at the command log. Also by using --log argument, you can have additional information.')

            patho_time = time.time()
            times.append(patho_time)
            steps.append('PathoLogic inference')
            logger.info('----------End of PathoLogic inference: {0:.2f}s----------'.format(times[-1] - times[-2]))
        else:
            multiprocess_run_pwt_flats = []
            multiprocess_run_move_pgdbs = []
            passed_inferences = []

    # Create path for lisp if there is no folder given.
    # Create the input for the creation of BioPAX/attribute-values files.
    if (flat_creation and not input_folder) or (output_folder and not input_folder):
        # Create a temporary folder in ptools-local where lisp script will be stored.
        tmp_folder = os.path.join(ptools_local_path, 'tmp')
        if not os.path.exists(tmp_folder):
            os.mkdir(tmp_folder)

        # Create a lisp script file for each PGDB in the ptools-local folder.
        flat_run_ids = create_only_flat_lisp(pgdbs_folder_path, tmp_folder)
        multiprocess_run_pwt_flats = []
        multiprocess_run_move_pgdbs = []
        for flat_run_id in flat_run_ids:
            input_tmp_folder_path = os.path.join(tmp_folder, flat_run_id)
            species_pgdb_folder = os.path.join(pgdbs_folder_path, flat_run_id.lower() + 'cyc')
            input_run_move_pgdbs = [flat_run_id, species_pgdb_folder]

            input_run_move_pgdbs.extend([dat_extraction, output_folder, size_reduction, xml_extraction, owl_extraction, col_extraction])
            multiprocess_run_pwt_flats.append([input_tmp_folder_path])
            multiprocess_run_move_pgdbs.append(input_run_move_pgdbs)

    # Add species that have data in PGDB but are not present in output folder.
    # Or if ignore_error has been used, select only PathoLogic build that have succeed + species in input with PGDB and not in output.
    if input_folder:
        if ignore_error:
            multiprocess_run_pwt_flats = []
            multiprocess_run_move_pgdbs = []
            if run_patho_flat_ids:
                if passed_inferences:
                    tmp_run_flat_ids = list(set(passed_inferences).intersection(set(run_patho_flat_ids)))
                else:
                    tmp_run_flat_ids = []
            else:
                tmp_run_flat_ids = []
            if run_flat_ids:
                tmp_run_flat_ids.extend(run_flat_ids)
            run_flat_ids = tmp_run_flat_ids
        if run_flat_ids:
            for run_flat_id in run_flat_ids:
                flat_creation_path = os.path.join(*[input_folder, run_flat_id, 'flat_files_creation.lisp'])
                create_flat_creation_script(run_flat_id, flat_creation_path)
                input_folder_path = os.path.join(input_folder, run_flat_id)
                species_pgdb_folder = os.path.join(pgdbs_folder_path, run_flat_id.lower() + 'cyc')
                multiprocess_run_pwt_flats.append([input_folder_path])
                input_run_move_pgdbs = [run_flat_id, species_pgdb_folder]
                input_run_move_pgdbs.extend([dat_extraction, output_folder, size_reduction, xml_extraction, owl_extraction, col_extraction])
                multiprocess_run_move_pgdbs.append(input_run_move_pgdbs)

    if not multiprocess_run_pwt_flats:
        close_mpwt(mpwt_pool, no_download_articles, pathway_score)
        logger.critical('No PGDB to export to move to output folder.')
        return

    if not multiprocess_run_move_pgdbs:
        close_mpwt(mpwt_pool, no_download_articles, pathway_score)
        logger.critical('No PGDB to export in flat files or to move to output folder.')
        return

    # Create BioPAX/attributes-values flat files.
    if (input_folder and flat_creation) or flat_creation:
        logger.info('~~~~~~~~~~Creation of the flat files~~~~~~~~~~')
        flat_error_status = mpwt_pool.starmap(run_pwt_flat, multiprocess_run_pwt_flats)
        logger.info('~~~~~~~~~~Check .dat~~~~~~~~~~')
        for multiprocess_run_move_pgdb in multiprocess_run_move_pgdbs:
            check_dat(multiprocess_run_move_pgdb[0], multiprocess_run_move_pgdb[1])
        if any(flat_error_status):
            if ignore_error:
                logger.critical('Error during flat files creation. Process stopped. Look at the command log. Also by using --log argument, you can have additional information.')
            else:
                close_mpwt(mpwt_pool, no_download_articles, pathway_score)
                sys.exit('Error during flat files creation. Process stopped. Look at the command log. Also by using --log argument, you can have additional information.')

        flat_time = time.time()
        times.append(flat_time)
        steps.append('BioPAX/attribute-value flat files creation')
        logger.info('----------End of flat files creation: {0:.2f}s----------'.format(times[-1] - times[-2]))

    if (flat_creation and not input_folder) or (output_folder and not input_folder):
        ptools_local_path = utils.find_ptools_path()
        ptools_local_tmp_path = os.path.join(ptools_local_path, 'tmp')
        shutil.rmtree(ptools_local_tmp_path)

    logger.info('~~~~~~~~~~End of Pathway Tools~~~~~~~~~~')

    # Move PGDBs or attribute-values/dat files.
    if output_folder:
        logger.info('~~~~~~~~~~Moving result files~~~~~~~~~~')
        if not os.path.exists(output_folder):
            logger.info('No output directory, it will be created.')
            os.mkdir(output_folder)
        mpwt_pool.starmap(run_move_pgdb, multiprocess_run_move_pgdbs)

        move_time = time.time()
        times.append(move_time)
        steps.append('Moving results files')
        logger.info('----------End of moving fimes: {0:.2f}s----------'.format(times[-1] - times[-2]))


    close_mpwt(mpwt_pool, no_download_articles, pathway_score)

    end_time = time.time()
    times.append(end_time)
    steps.append('mpwt')

    # Write each step time in log file.
    if patho_log:
        if patho_log:
            if not os.path.exists(patho_log):
                logger.info('No log directory, it will be created.')
                os.mkdir(patho_log)
        patho_error_pathname = os.path.join(patho_log, 'log_error.txt')
        with open(patho_error_pathname, 'a') as input_file:
            input_file.write('\n\n---------Time---------\n')
            for index, step_time in enumerate(times):
                if index != 0:
                    if index + 1 == len(times):
                        step_duration = step_time - times[0]
                    else:
                        step_duration = step_time - times[index-1]
                    input_file.write('Step {0} takes: {1:.2f}s.\n'.format(steps[index] , step_duration))

    logger.info('----------mpwt has finished in {0:.2f}s! Thank you for using it.'.format(end_time - start_time))


def close_mpwt(mpwt_pool, no_download_articles, pathway_score):
    """End multiprocessing Pool and restore ptools-init.dat

    mpwt_pool (multiprocessing Pool): mpwt multiprocessing Pool
    no_download_articles (bool): turning off loading of PubMed citations (True/False)
    pathway_score (float): score between 0 and 1 to accept or reject pathway
    """
    mpwt_pool.close()
    mpwt_pool.join()

    # Turn on loading of pubmed entries.
    if no_download_articles:
        utils.pubmed_citations(activate_citations=True)

    # Remodify the pathway score to its original value.
    if pathway_score:
        utils.modify_pathway_score(0.35)



"""
Scripts to convert GenBank files and GFF files into PathoLogic Files.
"""

import csv
import gffutils
import logging
import os
import shutil
import sys

from Bio import SeqIO
from multiprocessing import Pool

logger = logging.getLogger(__name__)


def create_pathologic_file(input_folder, output_folder, number_cpu=None):
    """
    Create PathoLogic file from Genbank or GFF files.

    Args:
        input_folder (str): pathname to the folder containing Genbanks or GFFs
        output_folder (str): pathname to the output folder containing the PathoLogic files
        number_cpu (str): number of CPU
    """
    if number_cpu:
        number_cpu_to_use = int(number_cpu)
    else:
        number_cpu_to_use = 1

    multiprocessing_input_data = []

    mpwt_pool = Pool(processes=number_cpu_to_use)

    if not os.path.exists(input_folder):
        logger.critical('mpwt can not run: ' + input_folder + ' does not exist.')
        return
    if not os.path.isdir(input_folder):
        logger.critical('mpwt can not run: ' + input_folder + ' is not a directory.')
        return

    input_names = os.listdir(input_folder)

    # If there is a taxon_id file in input read it and extract informations.
    if 'taxon_id.tsv' in input_names:
        taxon_ids = {}
        input_names.remove('taxon_id.tsv')
        taxon_id_path = os.path.join(input_folder, 'taxon_id.tsv')
        with open(taxon_id_path, 'r') as taxon_file:
            for row in csv.reader(taxon_file, delimiter='\t'):
                taxon_ids[row[0]] = row[1]
    else:
        taxon_ids = None

    if not os.path.exists(output_folder):
        os.makedirs(output_folder)

    # Initiate the output taxon_id.tsv file.
    output_taxon_id_path = os.path.join(output_folder, 'taxon_id.tsv')
    with open(output_taxon_id_path, 'w', encoding='utf-8') as taxon_id_file:
        taxon_writer = csv.writer(taxon_id_file, delimiter='\t')
        taxon_writer.writerow(['species', 'taxon_id'])

    # For each input, search for a GenBank, a GFF or PathoLogic files.
    for input_name in input_names:
        input_path_gbk = os.path.join(*[input_folder, input_name, input_name + '.gbk'])
        input_path_gbff = os.path.join(*[input_folder, input_name, input_name + '.gbff'])
        input_path_gff = os.path.join(*[input_folder, input_name, input_name + '.gff'])
        if os.path.exists(input_path_gbk):
            input_path = input_path_gbk
        elif os.path.exists(input_path_gbff):
            input_path = input_path_gbff
        elif os.path.exists(input_path_gff):
            input_path = input_path_gff
        elif any([species_file.endswith('.pf') for species_file in os.listdir(os.path.join(input_folder, input_name))]):
            input_path = os.path.join(input_folder, input_name)
        else:
            sys.exit('No .gff/.gbk/.gbff/.pf file in ' + os.path.join(input_folder, input_name))

        output_path = os.path.join(output_folder, input_name)

        if taxon_ids:
            if input_name in taxon_ids:
                taxon_id = taxon_ids[input_name]
            else:
                taxon_id = None
        else:
            taxon_id = None

        multiprocessing_input_data.append([input_path, output_path, output_folder, input_name, taxon_id])

    # Convert GenBank/GFF files to PathoLogic files.
    check_boolean = mpwt_pool.starmap(run_create_pathologic_file, multiprocessing_input_data)

    mpwt_pool.close()
    mpwt_pool.join()

    if not all(check_boolean):
        sys.exit('Issue during PathoLogic files creation, check errors.')


def write_taxon_id_file(input_name, taxon_id, output_folder):
    """
    Add species to taxon_id.tsv file.

    Args:
        input_name (str): name of the species
        taxon_id (str): taxon_id linked to the species
        output_folder (str): path to output folder
    """
    taxon_id_path = os.path.join(output_folder, 'taxon_id.tsv')

    with open(taxon_id_path, 'a') as taxon_id_file:
        taxon_writer = csv.writer(taxon_id_file, delimiter='\t')
        taxon_writer.writerow([input_name, taxon_id])


def run_create_pathologic_file(input_path, output_path, output_folder, input_name, taxon_id):
    """
    Create PathoLogic files from a Genbank or a GFF file.

    Args:
        input_path (str): path to species input folder
        output_path (str): path to output species folder
        output_folder (str): path to output folder
        input_name (str): species name
        taxon_id (dictionary): dictionary with the taxon_id for each species, if taxon_id.tsv does not exit None
    """
    # Add taxon ID in taxon_id.tsv if available.
    if input_path.endswith('.gbk') or input_path.endswith('.gbff'):
        logger.info('Creating PathoLogic file for ' + input_path)
        if not os.path.exists(output_path):
            os.makedirs(output_path)

        with open(input_path, "r") as gbk:
            first_seq_record = next(SeqIO.parse(gbk, "genbank"))
            src_features = [feature for feature in first_seq_record.features if feature.type == "source"]
            for src_feature in src_features:
                try:
                    src_dbxref_qualifiers = src_feature.qualifiers['db_xref']
                    for src_dbxref_qualifier in src_dbxref_qualifiers:
                        if 'taxon:' in src_dbxref_qualifier:
                            taxon_id = src_dbxref_qualifier.replace('taxon:', '')
                except KeyError:
                    logger.info('No taxon ID in the Genbank {0} In the FEATURES source you must have: /db_xref="taxon:taxonid" Where taxonid is the Id of your organism. You can find it on the NCBI.'.format(input_path))
            if taxon_id:
                write_taxon_id_file(input_name, taxon_id, output_folder)

        for record in SeqIO.parse(input_path, 'genbank'):
            element_id = record.id
            records = [record]
            output_fasta_path = os.path.join(output_path, element_id + '.fasta')
            output_pf_path = os.path.join(output_path, element_id + '.pf')
            SeqIO.write(records, output_fasta_path, 'fasta')
            with open(output_pf_path, 'w', encoding='utf-8') as element_file:
                element_file.write(';;;;;;;;;;;;;;;;;;;;;;;;;\n')
                element_file.write(';; ' + element_id + '\n')
                element_file.write(';;;;;;;;;;;;;;;;;;;;;;;;;\n')
                for feature in record.features:
                    if feature.type in ['rRNA', 'tRNA', 'ncRNA', 'misc_RNA', 'CDS']:
                        gene_name = None
                        gene_id = None
                        start_location = str(feature.location.start+1).replace('<', '').replace('>', '')
                        end_location = str(feature.location.end).replace('<', '').replace('>', '')
                        if 'locus_tag' in feature.qualifiers:
                            gene_id = feature.qualifiers['locus_tag'][0]
                        if 'gene' in feature.qualifiers:
                            gene_name = feature.qualifiers['gene'][0]
                        if not gene_id and not gene_name:
                            logger.critical('No locus_tag and no gene qualifiers in feature of record: ' + record.id + ' at position ' + start_location + '-' +end_location)
                            pass
                        if gene_name:
                            if len(gene_name) > 40:
                                logger.critical('Critical warning: gene ID ' + gene_id + ' of ' + feature.type + ' of file ' + input_path + 'is too long (more than 40 characters), this will cause errors in Pathway Tools.')
                            element_file.write('NAME\t' + gene_name + '\n')
                        else:
                            if gene_id:
                                if len(gene_id) > 40:
                                    logger.critical('Critical warning: gene ID ' + gene_id + ' of ' + feature.type + ' of file ' + input_path + 'is too long (more than 40 characters), this will cause errors in Pathway Tools.')
                                element_file.write('NAME\t' + gene_id + '\n')
                        if gene_id and gene_id != gene_name:
                            element_file.write('ID\t' + gene_id + '\n')
                        element_file.write('STARTBASE\t' + start_location + '\n')
                        element_file.write('ENDBASE\t' + end_location + '\n')
                        if 'product' in feature.qualifiers:
                            for function in feature.qualifiers['product']:
                                element_file.write('FUNCTION\t' + function + '\n')
                        if 'db_xref' in feature.qualifiers:
                            for db_xref in feature.qualifiers['db_xref']:
                                if ':' in db_xref:
                                    element_file.write('DBLINK\t' + db_xref + '\n')
                        if 'EC_number' in feature.qualifiers:
                            for ec in feature.qualifiers['EC_number']:
                                element_file.write('EC\t' + ec + '\n')
                        if 'go_component' in feature.qualifiers:
                            for go in feature.qualifiers['go_component']:
                                element_file.write('GO\t' + go + '\n')
                        if 'go_function' in feature.qualifiers:
                            for go in feature.qualifiers['go_function']:
                                element_file.write('GO\t' + go + '\n')
                        if 'go_process' in feature.qualifiers:
                            for go in feature.qualifiers['go_process']:
                                element_file.write('GO\t' + go + '\n')
                        if 'gene_synonym' in feature.qualifiers:
                            for gene_synonym in feature.qualifiers['gene_synonym']:
                                element_file.write('SYNONYM\t' + gene_synonym + '\n')
                        if feature.type == 'rRNA':
                            if 'pseudo' in feature.qualifiers:
                                element_file.write('PRODUCT-TYPE\tPSEUDO' + '\n')
                            else:
                                element_file.write('PRODUCT-TYPE\tRRNA' + '\n')
                            if gene_id:
                                element_file.write('PRODUCT-ID\trnra ' + gene_id + '\n')
                            else:
                                if gene_name:
                                    element_file.write('PRODUCT-ID\trnra ' + gene_name + '\n')
                            element_file.write('//\n\n')
                        if feature.type == 'misc_RNA' or feature.type == 'ncRNA':
                            if 'pseudo' in feature.qualifiers:
                                element_file.write('PRODUCT-TYPE\tPSEUDO' + '\n')
                            else:
                                element_file.write('PRODUCT-TYPE\tMISC-RNA' + '\n')
                            if gene_id:
                                element_file.write('PRODUCT-ID\tmiscnra ' + gene_id + '\n')
                            else:
                                if gene_name:
                                    element_file.write('PRODUCT-ID\tmiscnra ' + gene_name + '\n')
                            element_file.write('//\n\n')
                        if feature.type == 'tRNA':
                            if 'pseudo' in feature.qualifiers:
                                element_file.write('PRODUCT-TYPE\tPSEUDO' + '\n')
                            else:
                                element_file.write('PRODUCT-TYPE\tTRNA' + '\n')
                            if gene_id:
                                element_file.write('PRODUCT-ID\ttnra ' + gene_id + '\n')
                            else:
                                if gene_name:
                                    element_file.write('PRODUCT-ID\ttnra ' + gene_name + '\n')
                            element_file.write('//\n\n')
                        if feature.type == 'CDS':
                            if 'pseudo' in feature.qualifiers:
                                element_file.write('PRODUCT-TYPE\tPSEUDO' + '\n')
                            else:
                                element_file.write('PRODUCT-TYPE\tP' + '\n')
                            if gene_id:
                                element_file.write('PRODUCT-ID\tprot ' + gene_id + '\n')
                            else:
                                if gene_name:
                                    element_file.write('PRODUCT-ID\tprot ' + gene_name + '\n')
                            element_file.write('//\n\n')

                    elif feature.type == 'mRNA':
                         if 'pseudo' in feature.qualifiers:
                            start_location = str(feature.location.start+1).replace('<', '').replace('>', '')
                            end_location = str(feature.location.end).replace('<', '').replace('>', '')
                            if 'locus_tag' in feature.qualifiers:
                                gene_id = feature.qualifiers['locus_tag'][0]
                            if 'gene' in feature.qualifiers:
                                gene_name = feature.qualifiers['gene'][0]
                            if not gene_id and not gene_name:
                                logger.critical('No locus_tag and no gene qualifiers in feature of record: ' + record.id + ' at position ' + start_location + '-' +end_location)
                                pass
                            if gene_id:
                                if len(gene_id) > 40:
                                    logger.critical('Critical warning: gene ID ' + gene_id + ' of ' + feature.type + ' of file ' + input_path + 'is too long (more than 40 characters), this will cause errors in Pathway Tools.')
                                element_file.write('ID\t' + gene_id + '\n')
                            else:
                                if gene_name:
                                    if len(gene_name) > 40:
                                        logger.critical('Critical warning: gene ID ' + gene_id + ' of ' + feature.type + ' of file ' + input_path + 'is too long (more than 40 characters), this will cause errors in Pathway Tools.')
                                    element_file.write('ID\t' + gene_name + '\n')
                            if gene_name:
                                element_file.write('NAME\t' + gene_name + '\n')
                            else:
                                if gene_id:
                                    element_file.write('NAME\t' + gene_id + '\n')
                            element_file.write('STARTBASE\t' + start_location + '\n')
                            element_file.write('ENDBASE\t' + end_location + '\n')
                            element_file.write('PRODUCT-TYPE\tPSEUDO' + '\n')
                            if 'product' in feature.qualifiers:
                                for function in feature.qualifiers['product']:
                                    element_file.write('FUNCTION\t' + function + '\n')
                            if 'db_xref' in feature.qualifiers:
                                for db_xref in feature.qualifiers['db_xref']:
                                    if ':' in db_xref:
                                        element_file.write('DBLINK\t' + db_xref + '\n')
                            element_file.write('//\n\n')

    elif input_path.endswith('.gff'):

        if not os.path.exists(output_path):
            os.makedirs(output_path)

        gff_database = gffutils.create_db(input_path, ':memory:', force=True, keep_order=True, merge_strategy='merge', sort_attribute_values=True)
        regions = list(set([region.chrom for region in gff_database.features_of_type('region')]))
        try:
            first_region = list(set([region for region in gff_database.features_of_type('region')]))[0]
        except IndexError:
            raise IndexError("No 'region' feature in the GFF " + input_path)
        if 'Dbxref' in first_region.attributes:
            for dbxref in first_region.attributes['Dbxref']:
                if 'taxon' in dbxref:
                    taxon_id = dbxref.replace('taxon:', '')
        if taxon_id:
            write_taxon_id_file(input_name, taxon_id, output_folder)

        gff_fasta = None
        if os.path.exists(input_path.replace('.gff', '.fasta')):
            gff_fasta = input_path.replace('.gff', '.fasta')
        elif os.path.exists(input_path.replace('.gff', '.fsa')):
            gff_fasta = input_path.replace('.gff', '.fsa')
        if not gff_fasta:
            logger.critical('Error: No fasta file (.fasta or .fsa) for GFF file ' + input_path)
            return None

        for record in SeqIO.parse(gff_fasta, 'fasta'):
            gff_fasta_extension = os.path.splitext(gff_fasta)[1]
            output_fasta = os.path.join(output_path, record.id + gff_fasta_extension)
            SeqIO.write(record, output_fasta, 'fasta')

        for region in regions:
            region_pf_path = os.path.join(output_path, region + '.pf')
            with open(region_pf_path, 'w', encoding='utf-8') as element_file:
                element_file.write(';;;;;;;;;;;;;;;;;;;;;;;;;\n')
                element_file.write(';; ' + region + '\n')
                element_file.write(';;;;;;;;;;;;;;;;;;;;;;;;;\n')
                for feature in gff_database.features_of_type(tuple(gff_database.featuretypes())):
                    if feature.featuretype == 'gene':
                        if feature.chrom == region:
                            if len(feature.id) > 40:
                                logger.critical('Critical warning: gene ID ' + gene_id + ' of ' + input_path + 'is too long (more than 40 characters), this will cause errors in Pathway Tools.')
                            element_file.write('ID\t' + feature.id + '\n')
                            element_file.write('NAME\t' + feature.id + '\n')
                            element_file.write('STARTBASE\t' + str(feature.start) + '\n')
                            element_file.write('ENDBASE\t' + str(feature.stop) + '\n')
                            for child in gff_database.children(feature.id):
                                if 'product' in child.attributes:
                                    for product in child.attributes['product']:
                                        element_file.write('FUNCTION\t' + product + '\n')
                                if 'ec_number' in child.attributes:
                                    for ec in child.attributes['ec_number']:
                                        element_file.write('EC\t' + ec + '\n')
                                if 'db_xref' in child.attributes:
                                    if ':' in db_xref:
                                        element_file.write('DBLINK\t' + db_xref + '\n')
                                if child.featuretype == 'CDS':
                                    element_file.write('PRODUCT-TYPE\tP' + '\n')
                                    element_file.write('PRODUCT-ID\tprot ' + feature.id + '\n')
                                elif child.featuretype == 'tRNA':
                                    element_file.write('PRODUCT-TYPE\tTRNA' + '\n')
                                    element_file.write('PRODUCT-ID\ttrna ' + feature.id + '\n')
                                elif child.featuretype == 'rRNA':
                                    element_file.write('PRODUCT-TYPE\tRRNA' + '\n')
                                    element_file.write('PRODUCT-ID\rrna ' + feature.id + '\n')
                                elif child.featuretype == 'pseudogene':
                                    element_file.write('PRODUCT-TYPE\tPSEUDO' + '\n')

                            element_file.write('//\n\n')

    elif any([species_file.endswith('.pf') for species_file in os.listdir(input_path)]):
        write_taxon_id_file(input_name, taxon_id, output_folder)
        if not os.path.isdir(output_path):
            os.mkdir(output_path)
        for species_file in os.listdir(input_path):
            species_file_path = os.path.join(input_path, species_file)
            if species_file.endswith('.pf'):
                if os.path.exists(species_file_path.replace('.pf', '.fasta')):
                    fasta_file = species_file.replace('.pf', '.fasta')
                elif os.path.exists(species_file_path.replace('.pf', '.fsa')):

                    fasta_file = species_file.replace('.pf', '.fsa')
                else:
                    logger.critical('Error: No fasta file (.fasta or .fsa) for PathoLogic file ' + species_file_path)
                    return None
                output_species_path = os.path.join(output_path, species_file)
                input_fasta_path = os.path.join(input_path, fasta_file)
                output_fasta_path = os.path.join(output_path, fasta_file)
                shutil.copy2(species_file_path, output_species_path)
                shutil.copy2(input_fasta_path, output_fasta_path)

    return True


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Wrapping Pathway Tools for PathoLogic and attribute-values flat files creation.
Move results files to an output folder.
"""
import chardet
import logging
import os
import shutil
import signal
import subprocess
import sys

logger = logging.getLogger(__name__)


def pwt_check_error(species_input_folder_path, subprocess_stdout, cmd, error_status, subprocess_returncode=None, subprocess_stderr=None):
    """
    Print error messages when there is a subprocess error during PathoLogic run.

    Args:
        species_input_folder_path (str): pathname to the species input folder
        subprocess_stdout (str): stdout returns by subprocess
        cmd (str): command used which generates the error
        error_status (bool): error status of Pathway Tools
        subprocess_returncode (int): code returns by subprocess
        subprocess_stderr (str): stderr returns by subprocess
    Returns:
        boolean: True if there is an error during Pathway Tools run
    """
    if error_status:
        logger.critical('!!!!!!!!!!!!!!!!!----------------------------------------!!!!!!!!!!!!!!!!!')
    species_name = os.path.basename(species_input_folder_path)
    if subprocess_returncode:
        logger.critical('Error for {0} with PathoLogic subprocess, return code: {1}'.format(species_name, str(subprocess_returncode)))
    if subprocess_stderr:
        logger.critical('An error occurred :' + subprocess_stderr.decode('utf-8'))

    # Look for error in pathologic.log.
    if '-patho' in cmd:
        pathologic_erros = ['fatal error', 'Error']
        patho_error_status = check_log(species_input_folder_path, 'pathologic.log', error_status, pathologic_erros)
    else:
        patho_error_status = None

    if '-load' in cmd:
        load_errors = ['Error', 'fatal error', 'No protein-coding genes with sequence data found.', 'Cannot continue.']
        flat_error_status = check_log(species_input_folder_path, 'flat_files_creation.log', error_status, load_errors)
    else:
        flat_error_status = None

    error_status = any([error_status, patho_error_status, flat_error_status])

    if error_status:
        logger.critical('=== Pathway Tools log ===')
        for line in subprocess_stdout:
            if line != '':
                logger.critical('\t' + line)
        logger.critical('!!!!!!!!!!!!!!!!!----------------------------------------!!!!!!!!!!!!!!!!!')

    return error_status


def check_log(species_input_folder_path, log_filename, error_status, log_errors):
    """
    Look for error and fatal error in pathologic.log after build.

    Args:
        species_input_folder_path (str): pathname to the species input folder
        log_filename (str): name of the log file
        error_status (bool): True if there is an error during Pathway Tools run
        log_errors (list): list of strings containing possible errors.
    Returns:
        boolean: True if there is an error during Pathway Tools run
    """
    fatal_error_index = None
    log_file_path = os.path.join(species_input_folder_path, log_filename)
    with open(log_file_path, 'r') as log_file:
        for index, line in enumerate(log_file):
            if line != '':
                if not line.startswith(';;;'):
                    if any(error in line for error in log_errors) and not fatal_error_index:
                        fatal_error_index = index
                        logger.critical('=== Error in {0} for {1}==='.format(log_filename, species_input_folder_path))
                        logger.critical('\t' + 'Error from the {0} file: {1}'.format(log_filename, log_file_path))
                        logger.critical('\t' + line)
                        error_status = True
                    if fatal_error_index:
                        if index > fatal_error_index:
                            logger.critical('\t' + line)

    return error_status


def run_pwt(species_input_folder_path, patho_hole_filler, patho_operon_predictor, patho_transporter_inference):
    """
    Create PGDB using files created during 'create_flats_and_lisp' ('organism-params.dat' and 'genetic-elements.dat').
    With verbose run check_output to retrieve the output of subprocess (and show when Pathway Tools has been killed).
    Otherwise send the output to the null device.
    Command used:
    pathway-tools -no-web-cel-overview -no-cel-overview -no-patch-download -disable-metadata-saving -nologfile -patho

    Args:
        species_input_folder_path (str): path to input folder
        patho_hole_filler (bool): boolean to use or not PathoLogic Hole Filler
        patho_operon_predictor (bool): boolean to use or not PathoLogic Operon Predictor
        patho_transporter_inference (bool): boolean to use or not PathoLogic Transport Inference Parser
    Returns:
        boolean: True if there is an error during Pathway Tools run
    """
    cmd_options = ['-no-web-cel-overview', '-no-cel-overview', '-no-patch-download', '-disable-metadata-saving', '-nologfile']

    cmd_pwt = ['pathway-tools', *cmd_options, '-patho', species_input_folder_path]

    if patho_hole_filler:
        cmd_pwt.append('-hole-filler')

    if patho_operon_predictor:
        cmd_pwt.append('-operon-predictor')

    if patho_transporter_inference:
        cmd_pwt.append('-tip')

    logger.info(' '.join(cmd_pwt))

    error_status = None
    errors = ['Restart actions (select using :continue):']
    patho_lines = []

    # Name of the file containing the log from Pathway Tools terminal.
    pwt_log = os.path.join(species_input_folder_path, 'pwt_terminal.log')

    try:
        # Launch Pathway Tools PathoLogic.
        # Use start_new_session to group process ID to kill this process and its childs (with os.killpg).
        patho_subprocess = subprocess.Popen(cmd_pwt, stdout=subprocess.PIPE, start_new_session=True, universal_newlines="")
        with open(pwt_log, 'w', encoding='utf-8') as  pwt_writer:
            for patho_line in iter(patho_subprocess.stdout.readline, b''):
                encoding = chardet.detect(patho_line)['encoding']
                patho_line = patho_line.decode(encoding, errors='replace')
                pwt_writer.write(patho_line)

                # An error occured, kill Pathway Tools.
                if any(error in patho_line for error in errors):
                    logger.info('Error possibly with the genbank file.')
                    error_status = True
                    patho_subprocess.kill()
                    os.killpg(os.getpgid(patho_subprocess.pid), signal.SIGKILL)

                patho_lines.append(patho_line)
                patho_subprocess.poll()
                return_code = patho_subprocess.returncode

                # Check if Pathway Tools has been killed with returncode.
                # Also check if Pathway Tools has finished PathoLogic inference (returncode 0).
                if return_code or return_code == 0:
                    if return_code == 0:
                        patho_subprocess.stdout.close()
                        return error_status
                    elif return_code != 0:
                        raise subprocess.CalledProcessError(return_code, cmd_pwt)

    except subprocess.CalledProcessError as subprocess_error:
        # Check error with subprocess (when process is killed).
        error_status = True
        error_status = pwt_check_error(species_input_folder_path, patho_lines, cmd_pwt, error_status, subprocess_error.returncode, patho_subprocess.stderr)

    # Check pathologic.log for error.
    error_status = pwt_check_error(species_input_folder_path, patho_lines, cmd_pwt, error_status)

    patho_subprocess.stdout.close()

    return error_status


def run_pwt_flat(species_input_folder_path):
    """
    Create flat files using a lisp script created during 'create_flats_and_lisp'.
    Kill the subprocess when the command reach the Navigator Window opening proposition.
    If this proposition is not closed, the script can't continue.
    Command used:
    pathway-tools -no-patch-download -disable-metadata-saving -nologfile -load

    Args:
        species_input_folder_path (str): path to input folder
    Returns:
        boolean: True if there is an error during lisp script execution
    """
    lisp_path = os.path.join(species_input_folder_path, 'flat_files_creation.lisp')
    cmd_options = ['-no-patch-download', '-disable-metadata-saving', '-nologfile']
    cmd_flat = ['pathway-tools', *cmd_options, '-load', lisp_path]

    logger.info(' '.join(cmd_flat))

    error_status = None
    flat_creation_ends = ['Opening Navigator window.']
    load_errors = ['Error', 'fatal error', 'No protein-coding genes with sequence data found.', 'Cannot continue.']
    load_lines = []

    # Name of the file containing the log from Pathway Tools terminal.
    flat_log = os.path.join(species_input_folder_path, 'flat_files_creation.log')

    try:
        # Launch Pathway Tools lisp command.
        # Use start_new_session to group process ID to kill this process and its childs (with os.killpg).
        load_subprocess = subprocess.Popen(cmd_flat, stdout=subprocess.PIPE, start_new_session=True, universal_newlines="")
        with open(flat_log, 'w', encoding='utf-8') as flat_file_writer:
            for load_line in iter(load_subprocess.stdout.readline, b''):
                encoding = chardet.detect(load_line)['encoding']
                load_line = load_line.decode(encoding, errors='replace')
                flat_file_writer.write(load_line)

                # Lisp commnd has finished, kill Pathway Toosl trying to open navigator.
                if any(flat_end in load_line for flat_end in flat_creation_ends):
                    load_subprocess.stdout.close()
                    load_subprocess.kill()
                    os.killpg(os.getpgid(load_subprocess.pid), signal.SIGKILL)
                    return

                # There is an error, kill lisp command and return the error.
                if any(error in load_line for error in load_errors):
                    if not load_line.startswith(';;;'):
                        error_status = True
                        load_lines.append(load_line)
                        load_subprocess.kill()
                        os.killpg(os.getpgid(load_subprocess.pid), signal.SIGKILL)

                load_lines.append(load_line)
                load_subprocess.poll()
                return_code = load_subprocess.returncode
                if return_code:
                    raise subprocess.CalledProcessError(return_code, cmd_flat)

        # Check for error.
        error_status = pwt_check_error(species_input_folder_path, load_lines, cmd_flat, error_status)

    except subprocess.CalledProcessError as subprocess_error:
        # Check error with subprocess (when process is killed).
        error_status = pwt_check_error(species_input_folder_path, load_lines, cmd_flat, error_status, subprocess_error.returncode, load_subprocess.stderr)

    load_subprocess.stdout.close()

    return error_status


def run_move_pgdb(pgdb_folder_dbname, pgdb_folder_path, dat_extraction, output_folder, size_reduction, xml_extraction, owl_extraction, col_extraction):
    """
    Move the result files inside the shared folder containing the input data.
    pgdb_folder_dbname: ID of the species.
    pgdb_folder_path: path to the PGDB of the species (in ptools-local).

    Args:
        pgdb_folder_dbname (str): species ID
        pgdb_folder_path (str): path to species PGDB folder
        dat_extraction (bool): to extract or not the attribute-values files (.dat files)
        output_folder (str): path to output folder
        size_reduction (bool): to compress or not the results
        xml_extraction (bool): to extract or not the metabolic-reactions.xml'
        owl_extraction (bool): to extract or not the owl files
        col_extraction (bool): to extract or not the tabular files (.col files)
    """
    output_species = os.path.join(output_folder, pgdb_folder_dbname)

    keep_extensions = []
    if dat_extraction:
        keep_extensions.append('.dat')
    if xml_extraction:
        keep_extensions.append('.xml')
    if owl_extraction:
        keep_extensions.append('.owl')
    if col_extraction:
        keep_extensions.append('.col')

    if xml_extraction and len(keep_extensions) == 1:
        xml_file = os.path.join(*[pgdb_folder_path, '1.0', 'data', 'metabolic-reactions.xml'])
        if not os.path.exists(xml_file):
            logger.critical('Missing metabolic-reactions.xml for ' + pgdb_folder_dbname + '. To create it use mpwt --dat.')
            return
        shutil.copyfile(xml_file, output_species+'.xml')
        if size_reduction:
            shutil.rmtree(pgdb_folder_path)
        return

    if owl_extraction and len(keep_extensions) == 1:
        owl_level_2 = os.path.join(*[pgdb_folder_path, '1.0', 'data', 'biopax-level2.owl'])
        if not os.path.exists(owl_level_2):
            logger.critical('Missing biopax-level2.owl for ' + pgdb_folder_dbname + '. To create it use mpwt --dat.')
            return
        owl_level_3 = os.path.join(*[pgdb_folder_path, '1.0', 'data', 'biopax-level3.owl'])
        if not os.path.exists(owl_level_3):
            logger.critical('Missing biopax-level3.owl for ' + pgdb_folder_dbname + '. To create it use mpwt --dat.')
            return
        shutil.copyfile(owl_level_2, output_species+'-level2.owl')
        shutil.copyfile(owl_level_3, output_species+'-level3.owl')
        if size_reduction:
            shutil.rmtree(pgdb_folder_path)
        return

    if len(keep_extensions) > 0:
        pgdb_tmp_folder_path = os.path.join(*[pgdb_folder_path, '1.0', 'data'])
        if not os.path.exists(pgdb_tmp_folder_path):
            logger.critical('Missing ' + pgdb_tmp_folder_path + ' folder.')
            return
    else:
        pgdb_tmp_folder_path = pgdb_folder_path

    # If size_reduction, mpwt will create a compressed version of the PGDB in output folder.
    # It will also delete the PGDB folder in ptools-local.
    if size_reduction:
        if len(keep_extensions) > 0:
            for pgdb_file in os.listdir(pgdb_tmp_folder_path):
                pgdb_file_pathname = os.path.join(pgdb_tmp_folder_path, pgdb_file)
                pgdb_file_extension = os.path.splitext(pgdb_file)[1]
                if pgdb_file_extension not in keep_extensions:
                    if os.path.isfile(pgdb_file_pathname):
                        os.remove(pgdb_file_pathname)
                    elif os.path.isdir(pgdb_file_pathname):
                        shutil.rmtree(pgdb_file_pathname)
        zip_input_path = os.path.join(output_folder, pgdb_folder_dbname)
        shutil.make_archive(zip_input_path, 'zip', pgdb_tmp_folder_path)
        shutil.rmtree(pgdb_folder_path)

    else:
        shutil.copytree(pgdb_tmp_folder_path, output_species)
        if len(keep_extensions) > 0:
            for pgdb_file in os.listdir(output_species):
                pgdb_file_pathname = os.path.join(output_species, pgdb_file)
                pgdb_file_extension = os.path.splitext(pgdb_file)[1]
                if pgdb_file_extension not in keep_extensions:
                    if os.path.isfile(pgdb_file_pathname):
                        os.remove(pgdb_file_pathname)
                    elif os.path.isdir(pgdb_file_pathname):
                        shutil.rmtree(pgdb_file_pathname)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Description:
Test mpwt on genbank files containing E. coli genes implied in the TCA cycle and in the Fatty Acid Beta oxydation.
Need an environment with Pathway-Tools installed.
"""

import mpwt
import os
import shutil
import subprocess
import logging

logging.basicConfig(format='%(message)s', level=logging.CRITICAL)
logger = logging.getLogger(__name__)
logging.getLogger("mpwt").setLevel(logging.DEBUG)


def reaction_extraction(pathway_file_pathname):
    expected_reactions = []

    with open(pathway_file_pathname, 'r') as pathway_file:
        for line in pathway_file:
            if 'REACTION-LIST' in line:
                reaction = line.split(' - ')[1].strip()
                expected_reactions.append(reaction)

    return expected_reactions


def fabo_reactions():
    return ["ACYLCOASYN-RXN", "ACYLCOADEHYDROG-RXN", "ENOYL-COA-DELTA-ISOM-RXN", "ENOYL-COA-HYDRAT-RXN",
                    "OHBUTYRYL-COA-EPIM-RXN", "OHACYL-COA-DEHYDROG-RXN", "KETOACYLCOATHIOL-RXN"]


def test_multiprocess_pwt_import():
    """
    Test mpwt when called in a python script.
    """
    mpwt.remove_pgdbs('fatty_acid_beta_oxydation_icyc,fatty_acid_beta_oxydation_i_gffcyc,fatty_acid_beta_oxydation_i_pfcyc')
    mpwt.cleaning_input('test')

    mpwt.create_pathologic_file('test', 'test_pf')
    mpwt.multiprocess_pwt('test_pf', 'test_output', patho_inference=True, flat_creation=True, dat_extraction=True, size_reduction=False, number_cpu=3, verbose=True)

    pathway_fabo_pathname = os.path.join(*['test_output', 'fatty_acid_beta_oxydation_I_gff', 'pathways.dat'])
    expected_tca_reactions = reaction_extraction(pathway_fabo_pathname)
    assert set(fabo_reactions()).issubset(set(expected_tca_reactions))

    pathway_fabo_pathname = os.path.join(*['test_output', 'fatty_acid_beta_oxydation_I', 'pathways.dat'])
    expected_fabo_reactions = reaction_extraction(pathway_fabo_pathname)
    assert set(fabo_reactions()).issubset(set(expected_fabo_reactions))

    pathway_fabo_pathname = os.path.join(*['test_output', 'fatty_acid_beta_oxydation_I_pf', 'pathways.dat'])
    expected_pf_fabo_reactions = reaction_extraction(pathway_fabo_pathname)
    assert set(fabo_reactions()).issubset(set(expected_pf_fabo_reactions))

    mpwt.cleaning_input('test')
    shutil.rmtree('test_pf')
    shutil.rmtree('test_output')
    shutil.rmtree('__pycache__')


def test_multiprocess_pwt_call():
    """
    Test mpwt when called in terminal.
    """
    subprocess.call(['mpwt', '--delete', 'fatty_acid_beta_oxydation_icyc,fatty_acid_beta_oxydation_i_gffcyc,fatty_acid_beta_oxydation_i_pfcyc'])
    subprocess.call(['mpwt', '-f', 'test', '--clean'])
    subprocess.call(['mpwt', '-f', 'test', '--patho'])

    subprocess.call(['mpwt', '-o', 'test_output', '--flat', '--md', '--mx', '--mc', '--mo', '--cpu', '3'])

    pgdbs = mpwt.list_pgdb()
    assert set(['fatty_acid_beta_oxydation_i_gffcyc', 'fatty_acid_beta_oxydation_i_pfcyc', 'fatty_acid_beta_oxydation_icyc']).issubset(set(pgdbs))

    pathway_fabo_pathname = os.path.join(*['test_output', 'fatty_acid_beta_oxydation_i_gff', 'pathways.dat'])
    expected_tca_reactions = reaction_extraction(pathway_fabo_pathname)
    assert set(fabo_reactions()).issubset(set(expected_tca_reactions))

    pathway_fabo_pathname = os.path.join(*['test_output', 'fatty_acid_beta_oxydation_i', 'pathways.dat'])
    expected_fabo_reactions = reaction_extraction(pathway_fabo_pathname)
    assert set(fabo_reactions()).issubset(set(expected_fabo_reactions))

    pathway_fabo_pathname = os.path.join(*['test_output', 'fatty_acid_beta_oxydation_i_pf', 'pathways.dat'])
    expected_pf_fabo_reactions = reaction_extraction(pathway_fabo_pathname)
    assert set(fabo_reactions()).issubset(set(expected_pf_fabo_reactions))

    subprocess.call(['mpwt', '-f', 'test', '--clean'])
    shutil.rmtree('test_output')



from mpwt.pwt_wrapper import run_pwt, run_pwt_flat
from mpwt.mpwt_workflow import multiprocess_pwt
from mpwt.utils import cleaning, cleaning_input, find_ptools_path, list_pgdb, pubmed_citations, remove_pgdbs
from mpwt.to_pathologic import create_pathologic_file

__version__='0.6.0'


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Description:
From Genbank/GFF/PF files this script will create Pathway Tools input data, then run Pathway Tools's PathoLogic on them. It can also generate flat files.
The script takes a folder name as argument.

usage:
    mpwt -f=FOLDER [-o=FOLDER] [--patho] [--hf] [--op] [--tp] [--nc] [--flat] [--md] [--mx] [--mo] [--mc] [-p=FLOAT] [--cpu=INT] [-r] [-v] [--clean] [--log=FOLDER] [--ignore-error] [--taxon-file]
    mpwt --flat [-f=FOLDER] [-o=FOLDER] [--md] [--mx] [--mo] [--mc] [--cpu=INT] [-v]
    mpwt -o=FOLDER [--md] [--mx] [--mo] [--mc] [--cpu=INT] [-v]
    mpwt --clean [--cpu=INT] [-v]
    mpwt --delete=STR [--cpu=INT]
    mpwt --list
    mpwt --version
    mpwt topf -f=FOLDER -o=FOLDER [--cpu=INT] [--clean]

options:
    -h --help     Show help.
    -f=FOLDER     Working folder containing sub-folders with Genbank/GFF/PF files.
    -o=FOLDER    Output folder path. Will create a output folder in this folder.
    --patho    Will run an inference of Pathologic on the input files.
    --hf    Use with --patho. Run the Hole Filler using Blast.
    --op    Use with --patho. Run the Operon predictor of Pathway-Tools.
    --tp    Use with --patho. Run the Transport Inference Parser of Pathway-Tools.
    --nc    Use with --patho. Turn off loading of Pubmed entries.
    -p=FLOAT   Use with --patho. Modify PathoLogic pathway prediction score.
    --flat    Will create BioPAX/attribute-value flat files from PGDB.
    --md    Move the dat files into the output folder.
    --mx    Move the metabolic-reactions.xml file into the output folder.
    --mo    Move owl files into the output folder.
    --mc    Move tabular files into the output folder.
    --clean    Clean ptools-local folder, before any other operations.
    --delete=STR    Give a PGDB name and it will delete it (if multiple separe them with a ',', example: ecolicyc,athalianacyc).
    -r    Will delete files in ptools-local and compress results files to reduce results size (use it with -o).
    --cpu=INT     Number of cpu to use for the multiprocessing (default=1). [default: 1]
    --log=FOLDER     Create PathoLogic log files inside the given folder (use it with --patho).
    --list     List all PGDBs inside the ptools-local folder.
    --ignore-error     Ignore errors (PathoLogic and flat-files creation) and continue for successful builds.
    --taxon-file     For the use of the taxon_id.tsv file to find the taxon ID.
    -v     Verbose.
    --version     Version
    topf     Will convert Genbank and/or GFF files into PathoLogic Format file.

"""

import docopt
import logging
import os
import sys
import pkg_resources

from mpwt import utils, to_pathologic
from mpwt.mpwt_workflow import multiprocess_pwt
from multiprocessing import Pool

logging.basicConfig(format='%(message)s', level=logging.CRITICAL)
logger = logging.getLogger(__name__)
logging.getLogger("mpwt").setLevel(logging.CRITICAL)

VERSION = pkg_resources.get_distribution("mpwt").version
LICENSE = """Copyright (C) AuReMe
License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>.
Mpwt is free software: you are free to change and redistribute it.
There is NO WARRANTY, to the extent permitted by law.
"""

def run_mpwt():
    """
    Function used with a mpwt call in the terminal.
    """
    args = docopt.docopt(__doc__)

    input_folder = args['-f']
    output_folder = args['-o']
    patho_inference = args['--patho']
    patho_hole_filler = args['--hf']
    patho_operon_predictor = args['--op']
    patho_transporter_inference = args['--tp']
    no_download_articles = args['--nc']
    flat_creation = args['--flat']
    move_dat = args['--md']
    move_xml = args['--mx']
    move_owl = args['--mo']
    move_col = args['--mc']
    size_reduction = args['-r']
    number_cpu = args['--cpu']
    patho_log = args['--log']
    clean_arg = args['--clean']
    pgdb_to_deletes = args['--delete']
    pgdb_list = args['--list']
    ignore_error = args['--ignore-error']
    taxon_file = args['--taxon-file']
    pathway_score = args['-p']
    verbose = args['-v']
    topf = args['topf']
    version = args['--version']

    if version:
        print("Mpwt v" + VERSION  + "\n" + LICENSE)
        sys.exit()

    if verbose:
        logging.getLogger("mpwt").setLevel(logging.DEBUG)
        logger.setLevel(logging.DEBUG)

    if pgdb_list:
        pgdbs = utils.list_pgdb()
        if pgdbs == []:
            logger.critical('No PGDB inside ptools-local.')
        else:
            logger.critical(str(len(pgdbs)) + ' PGDB inside ptools-local:\n' + '\t'.join(pgdbs))
        sys.exit()

    # Delete PGDB if use of --delete argument.
    # Use a set to remove redudant PGDB.
    if pgdb_to_deletes:
        utils.remove_pgdbs(list(set(pgdb_to_deletes.split(','))), number_cpu)
        return

    if clean_arg:
        if verbose:
            logger.info('~~~~~~~~~~Remove local PGDB~~~~~~~~~~')

        if input_folder:
            utils.cleaning_input(input_folder, verbose)
            input_pgdb_to_deletes = [species.lower() + 'cyc' for species in os.listdir(input_folder) if not species.startswith('.') and species != 'taxon_id.tsv']
            utils.remove_pgdbs(input_pgdb_to_deletes, number_cpu)
        else:
            utils.cleaning(number_cpu, verbose)
        if not patho_inference and not flat_creation and not move_dat and not output_folder:
            sys.exit()

    if topf:
        if input_folder and output_folder:
            to_pathologic.create_pathologic_file(input_folder, output_folder, number_cpu)
        sys.exit()

    multiprocess_pwt(input_folder=input_folder,
                    output_folder=output_folder,
                    patho_inference=patho_inference,
                    patho_hole_filler=patho_hole_filler,
                    patho_operon_predictor=patho_operon_predictor,
                    patho_transporter_inference=patho_transporter_inference,
                    no_download_articles=no_download_articles,
                    flat_creation=flat_creation,
                    dat_extraction=move_dat,
                    xml_extraction=move_xml,
                    owl_extraction=move_owl,
                    col_extraction=move_col,
                    size_reduction=size_reduction,
                    number_cpu=number_cpu,
                    patho_log=patho_log,
                    ignore_error=ignore_error,
                    pathway_score=pathway_score,
                    taxon_file=taxon_file,
                    verbose=verbose)


if __name__ == "__main__":
    run_mpwt()



"""
Check results from Pathway Tools command:
-check_pwt: results from PathoLogic (by looking at pathologic.log)
-check_dat: attribute-values dat files
"""

import csv
import os
import logging
import sys

logger = logging.getLogger(__name__)


def check_pwt(multiprocess_run_pwts, patho_log_folder):
    """
    Check PathoLogic's log.
    Create two log files (log_error.txt which contains Pathway Tools log and resume_inference.tsv which contains summary of metabolic networks).

    Args:
        multiprocess_run_pwts (list): list of dictionary contaning multiprocess input data
        patho_log_folder (str): pathname to the PathoLogic log folder.

    Returns:
        list: Species with successful build.
    """
    if patho_log_folder:
        if not os.path.exists(patho_log_folder):
            logger.info('No log directory, it will be created.')
            os.mkdir(patho_log_folder)

        patho_error_pathname = os.path.join(patho_log_folder, 'log_error.txt')
        patho_resume_pathname = os.path.join(patho_log_folder, 'resume_inference.tsv')

        patho_error_file = open(patho_error_pathname, 'w', encoding='utf-8')
        patho_resume_file = open(patho_resume_pathname, 'w', encoding='utf-8')
        patho_resume_writer = csv.writer(patho_resume_file, delimiter='\t', lineterminator='\n')
        patho_resume_writer.writerow(['species', 'gene_number', 'protein_number', 'pathway_number', 'reaction_number', 'compound_number', 'pwt_non_fatal_error', 'pwt_warning'])

    failed_inferences = []
    passed_inferences = []

    for multiprocess_run_pwt in multiprocess_run_pwts:
        species_input_folder_path = multiprocess_run_pwt[0]
        species = os.path.basename(species_input_folder_path)
        patho_log = os.path.join(species_input_folder_path, 'pathologic.log')

        if patho_log_folder:
            patho_error_file.write('------------ Species: ')
            patho_error_file.write(species)
            patho_error_file.write('\n')

        fatal_error_index = None
        non_fatal_error_count = 0
        warning_count = 0

        if os.path.exists(patho_log):
            pgdb_build_done = False
            with open(patho_log, 'r') as input_file:
                for index, line in enumerate(input_file):
                    if ';;; Error:' in line:
                        non_fatal_error_count += 1
                    if 'Warning:' in line:
                        warning_count += 1
                    if 'fatal error' in line:
                        fatal_error_index = index
                        if species not in failed_inferences:
                            failed_inferences.append(species)
                        if patho_log_folder:
                            patho_error_file.write(line)
                            patho_resume_writer.writerow([species, 'ERROR', '', '', '', ''])

                    if fatal_error_index:
                        if index > fatal_error_index:
                            if patho_log_folder:
                                patho_error_file.write(line)

                    if 'Build done.' in line or 'PGDB build done.' in line:
                        pgdb_build_done = True
                        if patho_log_folder:
                            patho_error_file.write(line)
                            resume_inference_line = next(input_file)
                            patho_error_file.write(resume_inference_line)
                            if non_fatal_error_count > 0:
                                non_fatal_error_line = 'Number of non fatal errors: ' + str(non_fatal_error_count) + '. More information in ' + patho_log + '.\n'
                                patho_error_file.write(non_fatal_error_line)
                            if warning_count > 0:
                                warning_line = 'Number of warning: ' + str(warning_count) + '. More information in ' + patho_log + '.\n'
                                patho_error_file.write(warning_line)
                            gene_number = int(resume_inference_line.split('PGDB contains ')[1].split(' genes')[0])
                            protein_number = int(resume_inference_line.split('genes, ')[1].split(' proteins')[0])
                            pathway_number = int(resume_inference_line.split('proteins, ')[1].split(' base pathways')[0])
                            reaction_number = int(resume_inference_line.split('base pathways, ')[1].split(' reactions')[0])
                            compound_number = int(resume_inference_line.split('reactions, ')[1].split(' compounds')[0])

                    if 'Done' in line:
                        passed_inferences.append(species)
                        if patho_log_folder and pgdb_build_done:
                            patho_resume_writer.writerow([species, gene_number, protein_number, pathway_number, reaction_number, compound_number, non_fatal_error_count, warning_count])

                if species not in passed_inferences and species not in failed_inferences:
                    failed_inferences.append(species)
                    if patho_log_folder:
                        patho_error_file.write('No build in PathoLogic inference.')
                        patho_resume_writer.writerow([species, 'ERROR', '', '', '', ''])
        else:
            if patho_log_folder:
                patho_error_file.write('No pathologic log, an error occured before PathoLogic run.\n')
                patho_resume_writer.writerow([species, 'ERROR', '', '', '', ''])
            logger.info('No pathologic log for {0}, an error occured before PathoLogic run.'.format(species))

        if patho_log_folder:
            patho_error_file.write('------------\n\n')

    number_passed_inference = len(passed_inferences)
    number_failed_inference = len(failed_inferences)

    string_passed_build = 'build has' if number_passed_inference == 1 else 'builds have'
    string_failed_build = 'build has' if number_failed_inference == 1 else 'builds have'

    if number_passed_inference > 0:
        logger.info('\n{0} {1} passed!\n'.format(str(number_passed_inference), string_passed_build))
    if number_failed_inference > 0:
        logger.critical('WARNING: {0} {1} failed! See the log for more information.\n'.format(str(number_failed_inference), string_failed_build))

    if patho_log_folder:
        patho_error_file.close()
        patho_resume_file.close()
        with open(patho_error_pathname,'r') as contents:
            save = contents.read()
        with open(patho_error_pathname, 'w', encoding='utf-8') as output_file:
                output_file.write('Inference statistics:\n')
                if number_passed_inference > 0:
                    output_file.write('Build done: ' + str(number_passed_inference) + '\n')
                    output_file.write('Species: ' + ', '.join(passed_inferences) +  '\n\n')
                if number_failed_inference > 0:
                    output_file.write('Build failed: ' + str(number_failed_inference) + '\n')
                    output_file.write('Species: ' + ', '.join(failed_inferences) + '\n\n')
                output_file.write(save)

    return passed_inferences

def check_dat(run_dat_id, species_pgdb_folder):
    """
    Check dats creation.

    Args:
        run_dat_id (str): species ID
        species_pgdb_folder (str): path to species PGDB folder
    """
    pgdb_folder_dbname = run_dat_id.lower() + 'cyc'

    dats_path = os.path.join(*[species_pgdb_folder, '1.0', 'data'])

    dat_files = ["classes.dat", "compound-links.dat", "compounds.dat", "dnabindsites.dat", "enzrxns.dat", "gene-links.dat", "genes.dat", "pathway-links.dat",
                "pathways.dat", "promoters.dat", "protein-features.dat", "protein-links.dat", "proteins.dat", "protligandcplxes.dat", "pubs.dat",
                "reaction-links.dat", "reactions.dat", "regulation.dat", "regulons.dat", "rnas.dat", "species.dat", "terminators.dat", "transunits.dat"]

    dat_checks = []
    for dat_file in dat_files:
        dat_file_path = os.path.join(dats_path, dat_file)
        if os.path.exists(dat_file_path):
            dat_checks.append(dat_file_path)

    expected_dat_number = str(len(dat_files))
    found_dat_number = str(len(dat_checks))
    logger.info('{0}: {1} out of {2} dat files created.'.format(pgdb_folder_dbname, found_dat_number, expected_dat_number))




"""
Uselful functions for mpwt.
"""

import logging
import os
import shutil
import sys

from multiprocessing import Pool

logger = logging.getLogger(__name__)


def find_ptools_path():
    """
    Check if Pathway Tools is in PATH. If not stop the script.
    If it finds it, return the path of ptools-local by reading Pathway Tools file.

    Args:
    Returns:
        string: pathname to ptools-local folder
    """
    pathway_tools_path = shutil.which('pathway-tools')
    if not pathway_tools_path:
        logger.critical('Pathway Tools is not in the Path, mpwt can not work without it.')
        sys.exit(1)

    pathway_tools_file = open(pathway_tools_path, 'r')
    ptools_local_str = [line for line in pathway_tools_file if 'PTOOLS_LOCAL_PATH' in line][0]
    ptools_local_path = os.path.join(ptools_local_str.split(';')[0].split('=')[1].replace('"', '').strip(' '), 'ptools-local')
    pathway_tools_file.close()

    return ptools_local_path


def check_ptools_local_pwt():
    ptools_path = find_ptools_path()

    error = None

    ptools_init_path = os.path.join(ptools_path, 'ptools-init.dat')

    if not os.path.exists(ptools_init_path):
        logger.critical('Missing ptools-init.dat file in ptools-local folder. Use "pathway-tools -config" to recreate it.')
        error = True

    return error


def list_pgdb():
    """
    List all the PGDB inside the ptools-local folder.
    Return a list of their IDs.

    Args:
    Returns:
        list: PGDB IDs inside ptools-local
    """
    ptools_local_path = find_ptools_path()
    pgdb_folder = os.path.join(*[ptools_local_path, 'pgdbs', 'user'])

    return [species_pgdb for species_pgdb in os.listdir(pgdb_folder) if 'cyc' in species_pgdb]


def delete_pgdb(pgdb_name):
    """
    Remove a specific PGDB.

    Args:
        pgdb_name (str): PGDB ID to delete
    """
    ptools_local_path = find_ptools_path()
    pgdb_path = os.path.join(*[ptools_local_path.replace('\n', ''), 'pgdbs', 'user', pgdb_name])
    if os.path.isdir(pgdb_path):
        shutil.rmtree(pgdb_path)
        logger.info('{0} (at {1}) has been removed.'.format(pgdb_name, pgdb_path))


def remove_pgdbs(to_delete_pgdbs, number_cpu=None):
    """
    Delete all PGDB inside to_delete_pgdbs using multiprocessing.
    Check if there is a Pool and if not spawn one.

    Args:
        to_delete_pgdbs (list): PGDB IDs to delete
        number_cpu (str): number of CPUs to use (default = 1)
    """
    # Use the number of cpu given by the user or all the cpu available.
    if number_cpu:
        number_cpu_to_use = int(number_cpu)
    else:
        number_cpu_to_use = 1
    mpwt_pool = Pool(processes=number_cpu_to_use)

    mpwt_pool.map(delete_pgdb, to_delete_pgdbs)

    mpwt_pool.close()
    mpwt_pool.join()


def cleaning(number_cpu=None, verbose=None):
    """
    Clean Pathway Tools PGDB's folder.
    The script will delete folders and files in ptools-local/pgdbs/user.

    Args:
        number_cpu (str): number of CPUs to use
        verbose (bool): verbose argument
    """
    if verbose:
        logger.setLevel(logging.DEBUG)

    ptools_local_path = find_ptools_path()
    file_path = os.path.join(*[ptools_local_path.replace('\n', ''), 'pgdbs', 'user'])

    pgdb_metadata_path = os.path.join(file_path, 'PGDB-METADATA.ocelot')
    if os.path.isfile(pgdb_metadata_path):
        os.remove(pgdb_metadata_path)
        logger.info('PGDB-METADATA.ocelot has been removed.')

    pgdb_counter_path = os.path.join(file_path, 'PGDB-counter.dat')
    if os.path.isfile(pgdb_counter_path):
        os.remove(pgdb_counter_path)
        logger.info('PGDB-counter.dat has been removed.')

    # Extract all pgdbs inside ptools-local. Then delete them.
    all_pgdbs = os.listdir(file_path)
    remove_pgdbs(all_pgdbs, number_cpu)


def cleaning_input(input_folder, verbose=None):
    """
    Remove flat_files_creation.lisp, pathologic.log, genetic-elements.dat and organism-params.dat in a genbank folder.

    Args:
        input_folder (str): pathname to input folder
        verbose (bool): verbose argument
    """
    if verbose:
        logger.setLevel(logging.DEBUG)

    if not os.path.exists(input_folder):
        sys.exit('mpwt can not run: ' + input_folder + ' does not exist.')
    if not os.path.isdir(input_folder):
        sys.exit('mpwt can not run: ' + input_folder + ' is not a directory.')

    run_ids = [folder_id for folder_id in next(os.walk(input_folder))[1]]

    input_paths = [os.path.join(input_folder, run_id) for run_id in run_ids]

    for input_path in input_paths:
        if os.path.isdir(input_path):
            lisp_script = os.path.join(input_path, 'flat_files_creation.lisp')
            patho_log = os.path.join(input_path, 'pathologic.log')
            pwt_log = os.path.join(input_path, 'pwt_terminal.log')
            flat_log = os.path.join(input_path, 'flat_files_creation.log')
            genetic_dat = os.path.join(input_path, 'genetic-elements.dat')
            organism_dat = os.path.join(input_path, 'organism-params.dat')
            if os.path.exists(lisp_script):
                os.remove(lisp_script)
            if os.path.exists(patho_log):
                os.remove(patho_log)
            if os.path.exists(pwt_log):
                os.remove(pwt_log)
            if os.path.exists(flat_log):
                os.remove(flat_log)
            if os.path.exists(genetic_dat):
                os.remove(genetic_dat)
            if os.path.exists(organism_dat):
                os.remove(organism_dat)
            species = os.path.basename(input_path)
            logger.info('Remove ' + species + ' temporary datas.')


def pubmed_citations(activate_citations):
    """
    Activate or deactivate loading of PubMed citations.

    Args:
        activate_citations (bool): boolean to indicate if you want to activate or not the downlaod of Pubmed entries.
    """
    ptools_init_filepath = os.path.join(find_ptools_path(), 'ptools-init.dat')
    new_ptools_file = ""

    download_pubmed_entries_parameter = None
    with open(ptools_init_filepath, 'r') as ptools_init_file:
        for line in ptools_init_file.read().split('\n'):
            if 'Batch-PathoLogic-Download-Pubmed-Entries?' in line:
                if '#' in line:
                    line = line.replace('#', '')
                download_pubmed_entries_parameter = True
                if activate_citations:
                    line = line.replace('NIL', 'T')
                else:
                    line = line.replace('T', 'NIL')
            if line != '':
                new_ptools_file = new_ptools_file + line + '\n'
            else:
                new_ptools_file = new_ptools_file + line

    if not download_pubmed_entries_parameter:
        sys.exit('There is no Batch-PathoLogic-Download-Pubmed-Entries parameter in ' + ptools_init_filepath +'. To use --nc/no_download_articles, mpwt needs Pathway Tools 23.5 or higher.')

    with open(ptools_init_filepath, 'w', encoding='utf-8') as ptools_init_file:
        ptools_init_file.write(new_ptools_file)


def modify_pathway_score(pathway_score):
    """
    Modify the Pathway-Prediction-Score-Cutoff of ptools-init.dat

    Args:
        pathway_score (float): score between 0 and 1 to accept or reject pathways
    """
    ptools_init_filepath = os.path.join(find_ptools_path() ,'ptools-init.dat')
    new_ptools_file = ""

    pathway_prediction_score_cutoff = None
    with open(ptools_init_filepath, 'r') as ptools_init_file:
        for line in ptools_init_file.read().split('\n'):
            if 'Pathway-Prediction-Score-Cutoff' in line:
                pathway_prediction_score_cutoff = True
                if '#' in line:
                    line = line.replace('#', '')
                if pathway_score:
                    if pathway_score == 0.35:
                        line = '###' + line.split(' ')[0] + ' ' + str(pathway_score)
                    else:
                        line = line.split(' ')[0] + ' ' + str(pathway_score)
            if line != '':
                new_ptools_file = new_ptools_file + line + '\n'
            else:
                new_ptools_file = new_ptools_file + line

    if not pathway_prediction_score_cutoff:
        sys.exit('There is no Pathway-Prediction-Score-Cutoff parameter in ' + ptools_init_filepath +'.')

    with open(ptools_init_filepath, 'w', encoding='utf-8') as ptools_init_file:
        ptools_init_file.write(new_ptools_file)


import os

from distutils.util import convert_path
from setuptools import setup, find_packages

setup_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(setup_directory, 'README.rst'), encoding='utf-8') as readme_file:
    long_description = readme_file.read()

init_data = {}
init_pathname = convert_path('mpwt/__init__.py')
with open(init_pathname) as init_file:
    exec(init_file.read(), init_data)

setup(name='mpwt',
      description='Multiprocessing for Pathway Tools',
      long_description=long_description,
      version=init_data['__version__'],
      url='https://github.com/AuReMe/mpwt',
      author='Arnaud Belcour',
      python_requires='>=3.6',
      classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Audience
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',

        # Environnement, OS, languages
        'Programming Language :: Python :: 3'
      ],
      packages=['mpwt'],
      install_requires=[
            'biopython>=1.70',
            'chardet>=3.0.4',
            'docopt>=0.6.2',
            'gffutils>=0.9',
      ],
      entry_points={
          'console_scripts': [
              'mpwt = mpwt.__main__:run_mpwt'
          ]
      },
)
